import os

import sys
from pathlib import Path
from unittest.mock import patch
import pytest

from importlinter import cli
from importlinter.application import output
import io

this_directory = Path(__file__).parent
assets_directory = this_directory / ".." / "assets"

testpackage_directory = assets_directory / "testpackage"
multipleroots_directory = assets_directory / "multipleroots"
unmatched_ignore_imports_directory = testpackage_directory / "unmatched_ignore_imports_alerting"
namespace_packages_directory = assets_directory / "namespacepackages"
cyclicpackage_directory = assets_directory / "cycles"

# Add namespace packages to Python path
sys.path.extend(
    [str(namespace_packages_directory / location) for location in ("locationone", "locationtwo")],
)


@pytest.mark.parametrize(
    "working_directory, config_filename, expected_result",
    (
        (testpackage_directory, None, cli.EXIT_STATUS_SUCCESS),
        (testpackage_directory, ".brokencontract.ini", cli.EXIT_STATUS_ERROR),
        (testpackage_directory, ".malformedcontract.ini", cli.EXIT_STATUS_ERROR),
        (testpackage_directory, ".customkeptcontract.ini", cli.EXIT_STATUS_SUCCESS),
        (testpackage_directory, ".externalbrokencontract.ini", cli.EXIT_STATUS_ERROR),
        (
            multipleroots_directory,
            ".multiplerootskeptcontract.ini",
            cli.EXIT_STATUS_SUCCESS,
        ),
        (
            multipleroots_directory,
            ".multiplerootsbrokencontract.ini",
            cli.EXIT_STATUS_ERROR,
        ),
        (testpackage_directory, ".setup.toml", cli.EXIT_STATUS_SUCCESS),
        (testpackage_directory, ".customkeptcontract.toml", cli.EXIT_STATUS_SUCCESS),
        (testpackage_directory, ".externalkeptcontract.ini", cli.EXIT_STATUS_SUCCESS),
        (testpackage_directory, ".externalkeptcontract.toml", cli.EXIT_STATUS_SUCCESS),
        # Unmatched ignore imports alerting.
        # The return value depends on what this is set to.
        (
            testpackage_directory,
            str(unmatched_ignore_imports_directory / "unspecified.ini"),
            cli.EXIT_STATUS_ERROR,
        ),
        (
            testpackage_directory,
            str(unmatched_ignore_imports_directory / "error.ini"),
            cli.EXIT_STATUS_ERROR,
        ),
        (
            testpackage_directory,
            str(unmatched_ignore_imports_directory / "warn.ini"),
            cli.EXIT_STATUS_SUCCESS,
        ),
        (
            testpackage_directory,
            str(unmatched_ignore_imports_directory / "none.ini"),
            cli.EXIT_STATUS_SUCCESS,
        ),
        # Namespace packages
        (namespace_packages_directory, "keptcontract_portion.ini", cli.EXIT_STATUS_SUCCESS),
        (namespace_packages_directory, "keptcontract_namespace.ini", cli.EXIT_STATUS_SUCCESS),
        (namespace_packages_directory, "brokencontract_portion.ini", cli.EXIT_STATUS_ERROR),
        (namespace_packages_directory, "brokencontract_namespace.ini", cli.EXIT_STATUS_ERROR),
        # Type checking imports
        (testpackage_directory, ".typecheckkeptcontract.ini", cli.EXIT_STATUS_SUCCESS),
        (testpackage_directory, ".typecheckbrokencontract.ini", cli.EXIT_STATUS_ERROR),
        # Acyclic siblings
        (cyclicpackage_directory, ".keptcontract.ini", cli.EXIT_STATUS_SUCCESS),
        (cyclicpackage_directory, ".brokencontract.ini", cli.EXIT_STATUS_ERROR),
    ),
)
def test_lint_imports(working_directory, config_filename, expected_result):
    os.chdir(working_directory)

    if config_filename:
        result = cli.lint_imports(config_filename=config_filename)
    else:
        result = cli.lint_imports()

    assert expected_result == result


@pytest.mark.parametrize("is_debug_mode", (True, False))
def test_lint_imports_debug_mode(is_debug_mode):
    os.chdir(testpackage_directory)

    kwargs = dict(config_filename=".nonexistentcontract.ini", is_debug_mode=is_debug_mode)
    if is_debug_mode:
        with pytest.raises(FileNotFoundError):
            cli.lint_imports(**kwargs)
    else:
        assert cli.EXIT_STATUS_ERROR == cli.lint_imports(**kwargs)


def test_show_timings_smoke_test():
    os.chdir(testpackage_directory)
    assert cli.EXIT_STATUS_SUCCESS == cli.lint_imports(show_timings=True)


@pytest.mark.parametrize("verbose", (True, False))
def test_logging_configuration_respects_verbose_flag(verbose, capsys):
    os.chdir(testpackage_directory)

    cli.lint_imports(verbose=verbose)

    captured = capsys.readouterr()

    # N.B. "Wrote data cache file" is logged by Grimp.
    assert ("Wrote data cache file" in captured.out) == verbose


def test_windows_terminal_encoding_smoke_test():
    default_windows_terminal_encoding = "cp1252"
    console = output.Console(
        file=io.TextIOWrapper(
            io.BytesIO(),
            encoding=default_windows_terminal_encoding,
        )
    )
    os.chdir(testpackage_directory)

    with patch.object(output, "console", console):
        cli.lint_imports()



import pytest
from textwrap import dedent
from importlinter.application.output import console
from importlinter.contracts.protected import ProtectedContract
from grimp import ImportGraph


class TestProtectedContract:
    @staticmethod
    def _build_default_graph() -> ImportGraph:
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.bar",
            "mypackage.bar.allowed",
            "mypackage.bar.allowed.three.blue",
            "mypackage.bar.allowed.one",
            "mypackage.bar.allowed.two",
            "mypackage.bar.other_package",
            "mypackage.bar.other_package.one",
            "mypackage.bar.other_package.two",
            "mypackage.foo.protected",
            "mypackage.foo.protected.models",
            "mypackage.foo.protected.models.one",
            "mypackage.foo.protected.models.two",
            "mypackage.foo.protected.other_models",
            "mypackage.other_protected",
            "mypackage.other_protected.blue",
            "mypackage.foo.sibling",
            "mypackage.foo.sibling.models",
        ):
            graph.add_module(module)
        return graph

    @pytest.mark.parametrize(
        "import_details,contract_kept,description",
        [
            (
                {
                    "importer": "mypackage.bar.allowed",
                    "imported": "mypackage.foo.protected",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Allowed module can import protected",
            ),
            (
                {
                    "importer": "mypackage.foo.protected.other_models",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Modules in protected package can import each others",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.sibling.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Unrelated imports are still valid",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Allowed module can import children of protected",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.protected",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Other packages are not allowed to import protected",
            ),
            (
                {
                    "importer": "mypackage.foo.sibling",
                    "imported": "mypackage.foo.protected",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Siblings modules are not allowed to import protected",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed.three.blue",
                    "imported": "mypackage.foo.protected",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Descendants of allowed module are not allowed to import protected",
            ),
        ],
    )
    def test_detects_illegal_imports_correctly_without_as_package(
        self, import_details, contract_kept, description
    ):
        # Tested contract says that the module mypackage.foo.protected.models can be imported
        # only from mypackage.foo.protected children and mypackage.bar.allowed children

        graph = self._build_default_graph()
        graph.add_import(**import_details)

        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": ("mypackage.foo.protected"),
                "allowed_importers": ("mypackage.bar.allowed"),
                "as_packages": "False",
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)
        assert contract_check.kept == contract_kept, description

    @pytest.mark.parametrize(
        "import_details,contract_kept,description",
        [
            (
                {
                    "importer": "mypackage.foo.protected.other_models",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Modules inside protected package can import each other",
            ),
            (
                {
                    "importer": "mypackage.foo.sibling.models",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Modules inside siblings of protected package are not allowed to import it",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Modules not mentioned in the contract are not allowed to import protected ones",
            ),
            (
                {
                    "importer": "mypackage.other_protected.blue",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Other protected modules are not allowed to import protected ones",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.sibling.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Unrelated modules are still allowed to import each others",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed",
                    "imported": "mypackage.foo.protected",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Allowed package can import protected one at root level",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed.three.blue",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Deeper allowed modules are allowed to import protected ones",
            ),
        ],
    )
    def test_detects_illegal_imports_correctly_with_as_package(
        self, import_details, contract_kept, description
    ):
        # Tested contracts says the package mypackage.foo.protected can only be imported
        # from mypackage.bar.allowed
        graph = self._build_default_graph()
        graph.add_import(**import_details)

        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": [
                    "mypackage.foo.protected",
                    "mypackage.other_protected",
                ],
                "allowed_importers": ["mypackage.bar.allowed"],
                "as_packages": "True",
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)
        assert contract_check.kept == contract_kept, description

    def test_can_add_protected_modules_to_allowed_importers(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.other_protected.blue",
            imported="mypackage.foo.protected.models",
        )
        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": [
                    "mypackage.foo.protected",
                    "mypackage.other_protected",
                ],
                "allowed_importers": ["mypackage.other_protected"],
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept is True

    @pytest.mark.parametrize(
        "import_details,contract_kept,description",
        [
            (
                {
                    "importer": "mypackage.bar.allowed",
                    "imported": "mypackage.foo.protected",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Allowed module root can import protected root",
            ),
            (
                {
                    "importer": "mypackage.foo.protected.models.one",
                    "imported": "mypackage.foo.protected.models.two",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Modules in same protected top-level module can import each others",
            ),
            (
                {
                    "importer": "mypackage.foo.protected.other_models",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Modules in different protected top-level module cannot import each others",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.sibling.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                True,
                "Unrelated imports are still valid",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Allowed module expression root cannot import protected modules",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Other modules are not allowed to import protected",
            ),
            (
                {
                    "importer": "mypackage.foo.sibling",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Siblings of root are not allowed to import protected modules",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed.three.blue",
                    "imported": "mypackage.foo.protected.models",
                    "line_number": 3,
                    "line_contents": "print",
                },
                False,
                "Descendants of allowed module are not allowed to import protected one",
            ),
        ],
    )
    def test_detects_illegal_imports_if_target_uses_wildcards_but_not_as_packages(
        self, import_details, contract_kept, description
    ):
        graph = self._build_default_graph()
        graph.add_import(**import_details)

        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": ("mypackage.foo.protected.*"),
                "allowed_importers": ("mypackage.bar.allowed.*"),
                "as_packages": "False",
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)
        assert contract_check.kept == contract_kept, description

    def test_handle_duplicated_illegal_imports(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.bar.other_package.one",
            imported="mypackage.foo.protected.models",
            line_number=6,
            line_contents="import models",
        )
        graph.add_import(
            importer="mypackage.bar.other_package.one",
            imported="mypackage.foo.protected.models",
            line_number=27,
            line_contents="import models",
        )

        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": ("mypackage.foo.protected.*"),
                "allowed_importers": ("mypackage.bar.allowed.*"),
                "as_packages": "False",
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)
        assert not contract_check.kept

    @pytest.mark.parametrize(
        "import_details,as_packages,contract_kept,description",
        [
            (
                {
                    "importer": "mypackage.bar.allowed",
                    "imported": "django",
                },
                "False",
                True,
                "Allowed module can import external protected module",
            ),
            (
                {
                    "importer": "mypackage.bar.allowed.one",
                    "imported": "django.core",
                },
                "True",
                True,
                "Allowed package can import external protected package",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package",
                    "imported": "django",
                },
                "False",
                False,
                "Non-allowed module cannot import external protected module",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package",
                    "imported": "mypackage.foo.sibling",
                },
                "False",
                True,
                "Unrelated imports are not affected without as_package",
            ),
            (
                {
                    "importer": "mypackage.bar.other_package.one",
                    "imported": "mypackage.foo.sibling",
                },
                "True",
                True,
                "Unrelated imports are not affected with as_package",
            ),
        ],
    )
    def test_protect_external_package(
        self, import_details, as_packages, contract_kept, description
    ):
        graph = self._build_default_graph()
        graph.add_module("django", is_squashed=True)
        graph.add_import(**import_details)
        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": ("django",),
                "allowed_importers": ("mypackage.bar.allowed",),
                "as_packages": as_packages,
            },
        )

        contract_check = contract.check(graph, verbose=False)

        assert contract_check.kept == contract_kept, description

    def test_render_broken_contract_simple_with_package(self):
        graph = self._build_default_graph()

        graph.add_import(
            importer="mypackage.foo.sibling",
            imported="mypackage.foo.protected.models",
            line_number=6,
            line_contents="import models",
        )

        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": ("mypackage.foo.protected"),
                "allowed_importers": ("mypackage.bar.allowed"),
                "as_packages": "True",
            },
        )
        contract_check = contract.check(graph=graph, verbose=False)

        with console.capture() as capture:
            contract.render_broken_contract(contract_check)

        assert not contract_check.kept
        assert capture.get() == dedent(
            """Illegal imports of protected package mypackage.foo.protected:

- mypackage.foo.sibling -> mypackage.foo.protected.models (l.6)


        """
        )

    def test_render_broken_contract_full_with_package(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.green",
            "mypackage.green.one",
            "mypackage.green.two",
            "mypackage.green.three",
            "mypackage.green.four",
            "mypackage.green.five",
            "mypackage.blue",
            "mypackage.blue.models",
            "mypackage.blue.models.alpha",
            "mypackage.yellow",
            "mypackage.yellow.one",
            "mypackage.orange",
            "mypackage.orange.data",
            "mypackage.brown",
            "mypackage.brown.alpha",
        ):
            graph.add_module(module)

        graph.add_import(
            importer="mypackage.green.one",
            imported="mypackage.blue.models",
            line_number=7,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.green.three",
            imported="mypackage.blue.models",
            line_number=12,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.green.three",
            imported="mypackage.blue.models",
            line_number=34,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.green.five",
            imported="mypackage.blue.models.alpha",
            line_number=4,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.yellow.one",
            imported="mypackage.orange.models",
            line_number=16,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.yellow.one",
            imported="mypackage.orange.data",
            line_number=17,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.yellow.one",
            imported="mypackage.brown",
            line_number=18,
            line_contents="print",
        )

        graph.add_import(
            importer="mypackage.yellow.one",
            imported="mypackage.brown.alpha",
            line_number=19,
            line_contents="print",
        )

        contract = ProtectedContract(
            name="Protected contract",
            session_options={
                "root_packages": ["mypackage"],
            },
            contract_options={
                "protected_modules": (
                    "mypackage.**.models",
                    "mypackage.**.data",
                    "mypackage.brown",
                ),
                "allowed_importers": ("mypackage.colors.*",),
                "as_packages": "true",
            },
        )
        contract_check = contract.check(graph=graph, verbose=False)

        with console.capture() as capture:
            contract.render_broken_contract(contract_check)

        assert capture.get() == dedent(
            """Illegal imports of protected package mypackage.blue.models
(via mypackage.**.models expression):

- mypackage.green.one -> mypackage.blue.models (l.7)

- mypackage.green.three -> mypackage.blue.models (l.12, 34)

- mypackage.green.five -> mypackage.blue.models.alpha (l.4)

Illegal imports of protected package mypackage.orange.models
(via mypackage.**.models expression):

- mypackage.yellow.one -> mypackage.orange.models (l.16)

Illegal imports of protected package mypackage.orange.data
(via mypackage.**.data expression):

- mypackage.yellow.one -> mypackage.orange.data (l.17)

Illegal imports of protected package mypackage.brown:

- mypackage.yellow.one -> mypackage.brown (l.18)

- mypackage.yellow.one -> mypackage.brown.alpha (l.19)


        """
        )



import enum

import re
from typing import Any

import pytest

from importlinter.domain.fields import (
    BooleanField,
    IntegerField,
    EnumField,
    Field,
    ImportExpressionField,
    ListField,
    ModuleField,
    SetField,
    StringField,
    ValidationError,
)
from importlinter.domain.imports import ImportExpression, Module, ModuleExpression


def test_field_cannot_be_instantiated_with_default_and_required():
    class SomeField(Field):
        def parse(self, raw_data: str | list) -> str:
            raise NotImplementedError

    with pytest.raises(ValueError, match="A required field cannot also provide a default value."):
        SomeField(required=True, default="something")


class BaseFieldTest:
    field_class: type[Field] | None = None
    field_kwargs: dict[str, Any] = {}

    def test_field(self, raw_data, expected_value):
        field = self.field_class(**self.field_kwargs)

        if isinstance(expected_value, Exception):
            with pytest.raises(expected_value.__class__, match=re.escape(expected_value.message)):
                field.parse(raw_data) == expected_value
        else:
            assert field.parse(raw_data) == expected_value


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        ("Hello, world!", "Hello, world!"),
        (
            ["one", "two", "three"],
            ValidationError("Expected a single value, got multiple values."),
        ),
    ),
)
class TestStringField(BaseFieldTest):
    field_class = StringField


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        ("true", True),
        ("fAlSe", False),
        ("bananas", ValidationError("Could not parse a boolean from 'bananas'.")),
        (
            ["one", "two", "three"],
            ValidationError("Expected a single value, got multiple values."),
        ),
    ),
)
class TestBooleanField(BaseFieldTest):
    field_class = BooleanField


class TestIntegerField(BaseFieldTest):
    @pytest.mark.parametrize(
        "raw_data, minimum, expected_value",
        (
            ("1", None, 1),
            ("0", None, 0),
            ("-1", None, -1),
            ("44", 45, ValidationError("Must be >= 45.")),
            ("45", 45, 45),
            ("46", 45, 46),
            ("3.141", None, ValidationError("'3.141' is not an integer.")),
            ("bananas", None, ValidationError("'bananas' is not an integer.")),
            (
                ["1", "2", "3"],
                None,
                ValidationError("Expected a single value, got multiple values."),
            ),
        ),
    )
    def test_field(self, raw_data, minimum, expected_value):
        if minimum is None:
            field = IntegerField()
        else:
            field = IntegerField(minimum=minimum)

        if isinstance(expected_value, Exception):
            with pytest.raises(expected_value.__class__, match=re.escape(expected_value.message)):
                field.parse(raw_data)
        else:
            assert field.parse(raw_data) == expected_value


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        ("mypackage.foo.bar", Module("mypackage.foo.bar")),
        (
            ["one", "two", "three"],
            ValidationError("Expected a single value, got multiple values."),
        ),
        # TODO - test that it belongs in the root package.
    ),
)
class TestModuleField(BaseFieldTest):
    field_class = ModuleField


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        (
            "mypackage.foo -> mypackage.bar",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo"),
                imported=ModuleExpression("mypackage.bar"),
            ),
        ),
        (
            "my_package.foo   ->   my_package.bar",  # Extra whitespaces are supported.
            ImportExpression(
                importer=ModuleExpression("my_package.foo"),
                imported=ModuleExpression("my_package.bar"),
            ),
        ),
        (
            "my_package.foo -> my_package.foo_bar",  # Underscores are supported.
            ImportExpression(
                importer=ModuleExpression("my_package.foo"),
                imported=ModuleExpression("my_package.foo_bar"),
            ),
        ),
        # Wildcards
        # ---------
        (
            "mypackage.foo.* -> mypackage.bar",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo.*"),
                imported=ModuleExpression("mypackage.bar"),
            ),
        ),
        (
            "mypackage.foo.*.baz -> mypackage.bar",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo.*.baz"),
                imported=ModuleExpression("mypackage.bar"),
            ),
        ),
        (
            "mypackage.foo -> mypackage.bar.*",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo"),
                imported=ModuleExpression("mypackage.bar.*"),
            ),
        ),
        (
            "*.*.* -> mypackage.*.foo.*",
            ImportExpression(
                importer=ModuleExpression("*.*.*"),
                imported=ModuleExpression("mypackage.*.foo.*"),
            ),
        ),
        (
            "mypackage.foo.** -> mypackage.bar",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo.**"),
                imported=ModuleExpression("mypackage.bar"),
            ),
        ),
        (
            "mypackage.foo.**.baz -> mypackage.bar",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo.**.baz"),
                imported=ModuleExpression("mypackage.bar"),
            ),
        ),
        (
            "mypackage.foo -> mypackage.bar.**",
            ImportExpression(
                importer=ModuleExpression("mypackage.foo"),
                imported=ModuleExpression("mypackage.bar.**"),
            ),
        ),
        (
            "** -> mypackage.**.foo.*",
            ImportExpression(
                importer=ModuleExpression("**"),
                imported=ModuleExpression("mypackage.**.foo.*"),
            ),
        ),
        # Invalid expressions
        # -------------------
        (
            ["one", "two", "three"],
            ValidationError("Expected a single value, got multiple values."),
        ),
        (
            "mypackage.foo - mypackage.bar",
            ValidationError('Must be in the form "package.importer -> package.imported".'),
        ),
        (
            "mypackage.foo.bar* -> mypackage.bar",
            ValidationError("A wildcard can only replace a whole module."),
        ),
        (
            "mypackage.foo.b*z -> mypackage.bar",
            ValidationError("A wildcard can only replace a whole module."),
        ),
        (
            "mypackage.foo.bar** -> mypackage.bar",
            ValidationError("A wildcard can only replace a whole module."),
        ),
        (
            "mypackage.**.*.foo -> mypackage.bar",
            ValidationError("A recursive wildcard cannot be followed by a wildcard."),
        ),
        (
            "mypackage.**.**.foo -> mypackage.bar",
            ValidationError("A recursive wildcard cannot be followed by a wildcard."),
        ),
        (
            "mypackage.*.**.foo -> mypackage.bar",
            ValidationError("A wildcard cannot be followed by a recursive wildcard."),
        ),
        (
            "mypackage.foo.b**z -> mypackage.bar",
            ValidationError("A wildcard can only replace a whole module."),
        ),
    ),
)
class TestImportExpressionField(BaseFieldTest):
    field_class = ImportExpressionField


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        (
            ["mypackage.foo", "mypackage.bar"],
            [Module("mypackage.foo"), Module("mypackage.bar")],
        ),
        (
            ["mypackage.foo", "mypackage.foo"],
            [Module("mypackage.foo"), Module("mypackage.foo")],
        ),
        (["mypackage.foo", "    "], [Module("mypackage.foo")]),
        ("singlevalue", [Module("singlevalue")]),
    ),
)
class TestListField(BaseFieldTest):
    field_class = ListField
    field_kwargs = dict(subfield=ModuleField())


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        (
            ["mypackage.foo", "mypackage.bar"],
            {Module("mypackage.foo"), Module("mypackage.bar")},
        ),
        (["mypackage.foo", "mypackage.foo"], {Module("mypackage.foo")}),
        (["mypackage.foo", "    "], {Module("mypackage.foo")}),
        ("singlevalue", {Module("singlevalue")}),
    ),
)
class TestSetField(BaseFieldTest):
    field_class = SetField
    field_kwargs = dict(subfield=ModuleField())


class MyEnum(enum.Enum):
    RED = "red"
    GREEN = "green"
    blue = "blue"
    NONMATCHING_ORANGE = "orange"


@pytest.mark.parametrize(
    "raw_data, expected_value",
    (
        # Values:
        ("", MyEnum.RED),
        ("green", MyEnum.GREEN),
        # Lowercase attributes:
        ("blue", MyEnum.blue),
        # Values that don't match attributes:
        ("orange", MyEnum.NONMATCHING_ORANGE),
        # Trailing/leading spaces:
        (" ", MyEnum.RED),
        ("  green  ", MyEnum.GREEN),
        # Invalid choices:
        (
            "yellow",
            ValidationError(
                "Invalid value 'yellow': expected 'red', 'green', 'blue' or 'orange'."
            ),
        ),
        # Case sensitive:
        (
            "GREEN",
            ValidationError("Invalid value 'GREEN': expected 'red', 'green', 'blue' or 'orange'."),
        ),
    ),
)
class TestEnumField(BaseFieldTest):
    field_class = EnumField
    field_kwargs = dict(enum=MyEnum, default=MyEnum.RED)


class TestEnumFieldExtras:
    """
    Extra tests for the EnumField.
    """

    def test_works_with_no_default_provided(self):
        assert EnumField(MyEnum).parse("green") == MyEnum.GREEN

    @pytest.mark.parametrize("required", (True, False))
    def test_required_attribute_is_set_when_provided(self, required):
        assert EnumField(MyEnum, required=required).required is required

    def test_required_attribute_is_true_by_default(self):
        assert EnumField(MyEnum).required is True

    def test_validation_message_for_two_value_enum(self):
        class TwoValueEnum(enum.Enum):
            GREEN = "green"
            BLUE = "blue"

        with pytest.raises(
            ValidationError, match="Invalid value 'purple': expected 'green' or 'blue'."
        ):
            EnumField(TwoValueEnum).parse("purple")

    def test_requires_string_members(self):
        class InvalidEnum(enum.Enum):
            GREEN = "green"
            BLUE = 2

        with pytest.raises(
            TypeError,
            match="Unsupported Enum for EnumField: member values must all be strings.",
        ):
            EnumField(InvalidEnum)



import yaml

from importlinter.application.ports import filesystem as ports


class FakeFileSystem(ports.FileSystem):
    def __init__(
        self,
        contents: str | None = None,
        content_map: dict[str, str] | None = None,
        working_directory: str | None = None,
    ) -> None:
        """
        Files can be declared as existing in the file system in two different ways, either
        in a contents string (which is a quick way of defining a lot of files), or in content_map
        (which specifies the actual contents of a file in the file system). For a file to be
        treated as existing, it needs to be declared in at least one of these. If it isn't
        declared in content_map, the file will behave as an empty file.

        Args:
            contents: a string in the following format:

                /path/to/mypackage/
                    __init__.py
                    foo/
                        __init__.py
                        one.py
                        two/
                            __init__.py
                            green.py
                            blue.py

            content_map: A dictionary keyed with filenames, with values that are the contents.
                         If present in content_map, .read(filename) will return the string.
                {
                    '/path/to/foo/__init__.py': "from . import one",
                }

            working_directory: The path to be treated as the current working directory, e.g.
                               '/path/to/directory'.
        """
        self.contents = self._parse_contents(contents)
        self.content_map = content_map if content_map else {}
        self.working_directory = working_directory

    def join(self, *components: str) -> str:
        return "/".join(components)

    def _parse_contents(self, raw_contents: str | None):
        """
        Returns the raw contents parsed in the form:
            {
                '/path/to/mypackage': {
                    '__init__.py': None,
                    'foo': {
                        '__init__.py': None,
                        'one.py': None,
                        'two': {
                            '__init__.py': None,
                            'blue.py': None,
                            'green.py': None,
                        }
                    }
                }
            }
        """
        if raw_contents is None:
            return {}

        # Convert to yaml for ease of parsing.
        yamlified_lines = []
        raw_lines = [line for line in raw_contents.split("\n") if line.strip()]

        dedented_lines = self._dedent(raw_lines)

        for line in dedented_lines:
            trimmed_line = line.rstrip().rstrip("/")
            yamlified_line = trimmed_line + ":"
            yamlified_lines.append(yamlified_line)

        yamlified_string = "\n".join(yamlified_lines)

        return yaml.safe_load(yamlified_string)

    def _dedent(self, lines: list[str]) -> list[str]:
        """
        Dedent all lines by the same amount.
        """
        first_line = lines[0]
        first_line_indent = len(first_line) - len(first_line.lstrip())

        def dedented(line):
            return line[first_line_indent:]

        return list(map(dedented, lines))

    def read(self, file_name: str, encoding: str | None = None) -> str:
        if not self.exists(file_name):
            raise FileNotFoundError  # pragma: nocover
        try:
            file_contents = self.content_map[file_name]
        except KeyError:
            return ""
        raw_lines = [line for line in file_contents.split("\n") if line.strip()]
        dedented_lines = self._dedent(raw_lines)
        return "\n".join(dedented_lines)

    def exists(self, file_name: str) -> bool:
        # The file should exist if it's either declared in contents or in content_map.
        if file_name in self.content_map.keys():
            return True

        found_directory = None
        for directory in self.contents.keys():
            if file_name.startswith(directory):
                found_directory = directory
        if not found_directory:
            return False

        relative_file_name = file_name[len(found_directory) + 1 :]
        file_components = relative_file_name.split("/")

        contents = self.contents[found_directory]
        for component in file_components:
            try:
                contents = contents[component]
            except KeyError:
                return False
        return True

    def getcwd(self) -> str:
        if self.working_directory:
            return self.working_directory
        raise RuntimeError("No working directory specified.")  # pragma: nocover



from contextlib import contextmanager

import pytest

from importlinter.domain.imports import (
    DirectImport,
    ImportExpression,
    Module,
    ModuleExpression,
)


@contextmanager
def does_not_raise():
    yield


class TestModule:
    def test_object_representation(self):
        test_object = Module("new_module")
        assert repr(test_object) == "<Module: new_module>"

    @pytest.mark.parametrize(
        ("first_object", "second_object", "expected_bool"),
        [
            (Module("first"), Module("second"), False),
            (Module("same"), Module("same"), True),
            (Module("different"), "different", False),
        ],
    )
    def test_equal_magic_method(self, first_object, second_object, expected_bool):
        comparison_result = first_object == second_object
        assert comparison_result is expected_bool

    @pytest.mark.parametrize(
        ("module", "expected_parent", "exception"),
        [
            (Module("parent.child"), Module("parent"), does_not_raise()),
            (Module("child"), Module(""), pytest.raises(ValueError)),
        ],
    )
    def test_parent(self, module, expected_parent, exception):
        with exception:
            assert module.parent == expected_parent

    @pytest.mark.parametrize(
        ("child", "parent", "expected_bool"),
        [
            (Module("parent.child"), Module("parent"), True),
            (Module("grandparent.parent.child"), Module("grandparent"), False),
            (Module("first_child"), Module("second_child"), False),
        ],
    )
    def test_is_child_of(self, child, parent, expected_bool):
        assert child.is_child_of(parent) is expected_bool

    @pytest.mark.parametrize(
        ("candidate", "package", "expected_bool"),
        [
            (Module("somepackage"), Module("somepackage"), True),
            (Module("somepackage.foo"), Module("somepackage"), True),
            (Module("somepackage.foo.blue"), Module("somepackage"), True),
            (Module("somepackage.foo.blue"), Module("somepackage.foo"), True),
            (
                Module("somepackage.foo.blue.one.alpha"),
                Module("somepackage.foo.blue"),
                True,
            ),
            (Module("somepackage"), Module("somepackage.foo"), False),
            (Module("somepackage"), Module("differentpackage"), False),
            (Module("somepackage.foo"), Module("differentpackage"), False),
            (
                Module("somepackage.foo.blue.one.alpha"),
                Module("somepackage.foo.green"),
                False,
            ),
        ],
    )
    def test_is_in_package(self, candidate, package, expected_bool):
        assert candidate.is_in_package(package) is expected_bool


class TestDirectImport:
    def test_object_representation(self):
        test_object = DirectImport(
            importer=Module("mypackage.foo"),
            imported=Module("mypackage.bar"),
        )
        assert repr(test_object) == "<DirectImport: mypackage.foo -> mypackage.bar>"

    @pytest.mark.parametrize(
        ("test_object", "expected_string"),
        [
            (
                DirectImport(importer=Module("mypackage.foo"), imported=Module("mypackage.bar")),
                "mypackage.foo -> mypackage.bar",
            ),
            (
                DirectImport(
                    importer=Module("mypackage.foo"),
                    imported=Module("mypackage.bar"),
                    line_number=10,
                ),
                "mypackage.foo -> mypackage.bar (l. 10)",
            ),
        ],
    )
    def test_string_object_representation(self, test_object, expected_string):
        assert str(test_object) == expected_string


class TestModuleExpression:
    def test_object_representation(self):
        expression = ModuleExpression("mypackage.foo.**")
        assert repr(expression) == "<ModuleExpression: mypackage.foo.**>"


class TestImportExpression:
    def test_object_representation(self):
        test_object = ImportExpression(
            importer=ModuleExpression("mypackage.foo"),
            imported=ModuleExpression("mypackage.bar"),
        )
        assert repr(test_object) == "<ImportExpression: mypackage.foo -> mypackage.bar>"

    def test_string_object_representation(self):
        expression = ImportExpression(
            importer=ModuleExpression("mypackage.foo"),
            imported=ModuleExpression("mypackage.bar"),
        )
        assert str(expression) == "mypackage.foo -> mypackage.bar"

    @pytest.mark.parametrize(
        "first, second, expected",
        [
            (
                ImportExpression(
                    importer=ModuleExpression("mypackage.foo"),
                    imported=ModuleExpression("mypackage.bar"),
                ),
                ImportExpression(
                    importer=ModuleExpression("mypackage.foo"),
                    imported=ModuleExpression("mypackage.bar"),
                ),
                True,
            ),
            (
                ImportExpression(
                    importer=ModuleExpression("mypackage.foo"),
                    imported=ModuleExpression("mypackage.bar"),
                ),
                ImportExpression(
                    importer=ModuleExpression("mypackage.bar"),
                    imported=ModuleExpression("mypackage.foo"),
                ),
                False,
            ),
            (
                ImportExpression(
                    importer=ModuleExpression("mypackage.foo"),
                    imported=ModuleExpression("mypackage.bar"),
                ),
                ImportExpression(
                    importer=ModuleExpression("mypackage.foo"),
                    imported=ModuleExpression("mypackage.foobar"),
                ),
                False,
            ),
            (
                ImportExpression(
                    importer=ModuleExpression("mypackage.foo"),
                    imported=ModuleExpression("mypackage.bar"),
                ),
                ImportExpression(
                    importer=ModuleExpression("mypackage.foobar"),
                    imported=ModuleExpression("mypackage.bar"),
                ),
                False,
            ),
        ],
    )
    def test_equality(self, first, second, expected):
        assert expected == (first == second)

    @pytest.mark.parametrize(
        "importer, imported, has_wildcard_expression",
        [
            ("mypackage.foo", "mypackage.bar", False),
            ("mypackage.*", "mypackage.bar", True),
            ("mypackage.foo", "mypackage.*", True),
            ("mypackage.*", "mypackage.*", True),
            ("mypackage.*.foo", "mypackage.*.bar", True),
            ("mypackage.**", "mypackage.bar", True),
            ("mypackage.foo", "mypackage.**", True),
            ("mypackage.**", "mypackage.**", True),
            ("mypackage.**.foo", "mypackage.**.bar", True),
        ],
    )
    def test_has_wildcard_expression(self, importer, imported, has_wildcard_expression):
        expression = ImportExpression(
            importer=ModuleExpression(importer), imported=ModuleExpression(imported)
        )
        assert expression.has_wildcard_expression() == has_wildcard_expression



from importlinter.application.ports.user_options import UserOptionReader
from importlinter.application.user_options import UserOptions


class FakeUserOptionReader(UserOptionReader):
    def __init__(self, user_options: UserOptions):
        self._user_options = user_options

    def read_options(self, config_filename: str | None = None) -> UserOptions | None:
        return self._user_options


class ExceptionRaisingUserOptionReader(UserOptionReader):
    def __init__(self, exception: Exception):
        self._exception = exception

    def read_options(self, config_filename: str | None = None) -> UserOptions | None:
        raise self._exception



from __future__ import annotations
from typing import Any


class ValueObject:
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        else:
            return False

    def __hash__(self) -> int:
        return hash(str(self))


class Module(ValueObject):
    """
    A Python module.
    """

    def __init__(self, name: str) -> None:
        """
        Args:
            name: The fully qualified name of a Python module, e.g. 'package.foo.bar'.
        """
        self.name = name

    def __str__(self) -> str:
        return self.name

    @property
    def root_package_name(self) -> str:
        return self.name.split(".")[0]

    @property
    def parent(self) -> "Module":
        components = self.name.split(".")
        if len(components) == 1:
            raise ValueError("Module has no parent.")
        return Module(".".join(components[:-1]))

    def is_child_of(self, module: "Module") -> bool:
        try:
            return module == self.parent
        except ValueError:
            # If this module has no parent, then it cannot be a child of the supplied module.
            return False

    def is_descendant_of(self, module: "Module") -> bool:
        return self.name.startswith(f"{module.name}.")

    def is_in_package(self, package: "Module") -> bool:
        return self == package or self.is_descendant_of(package)


class DirectImport(ValueObject):
    """
    An import between one module and another.
    """

    def __init__(
        self,
        *,
        importer: Module,
        imported: Module,
        line_number: int | None = None,
        line_contents: str | None = None,
    ) -> None:
        self.importer = importer
        self.imported = imported
        self.line_number = line_number
        self.line_contents = line_contents

    def __str__(self) -> str:
        if self.line_number:
            return f"{self.importer} -> {self.imported} (l. {self.line_number})"
        else:
            return f"{self.importer} -> {self.imported}"

    def __hash__(self) -> int:
        return hash((str(self), self.line_contents))


class ModuleExpression(ValueObject):
    """
    A user-submitted expression describing a module or a set of modules.

    Sets of modules are notated using * or ** wildcards.
    Examples:
        "mypackage.foo.bar": a single module
        "mypackage.foo.*": all direct submodules in the foo subpackage
        "mypackage.*.bar": all bar-submodules of any mypackage submodule
        "mypackage.**": all modules in the mypackage package

    Note that * and ** cannot be mixed in the same expression.
    """

    def __init__(self, expression: str) -> None:
        self.expression = expression

    def has_wildcard_expression(self) -> bool:
        return "*" in self.expression

    def __str__(self) -> str:
        return self.expression

    def __lt__(self, other: ModuleExpression) -> bool:
        return self.expression < other.expression


class ImportExpression(ValueObject):
    """
    A user-submitted expression describing an import or set of imports.

    The importer and imported expressions are both ModuleExpressions
    (see ModuleExpression for details).
    """

    def __init__(self, importer: ModuleExpression, imported: ModuleExpression) -> None:
        self.importer = importer
        self.imported = imported

    def has_wildcard_expression(self) -> bool:
        return self.imported.has_wildcard_expression() or self.importer.has_wildcard_expression()

    def __str__(self) -> str:
        return f"{self.importer.expression} -> {self.imported.expression}"



import importlib
from copy import copy, deepcopy
from typing import Any
import contextlib
from grimp import ImportGraph
from rich.live import Live
from rich.progress import (
    Progress,
    BarColumn,
    MofNCompleteColumn,
    TextColumn,
)

from ..application import rendering
from ..domain.contract import Contract, InvalidContractOptions, registry
from .output import console
from . import output
from .app_config import settings
from .ports.reporting import Report
from .rendering import render_exception, render_report, format_duration
from .sentinels import NotSupplied
from .user_options import UserOptions

# Public functions
# ----------------

SUCCESS = True
FAILURE = False


def lint_imports(
    config_filename: str | None = None,
    limit_to_contracts: tuple[str, ...] = (),
    cache_dir: str | None | type[NotSupplied] = NotSupplied,
    is_debug_mode: bool = False,
    show_timings: bool = False,
    verbose: bool = False,
) -> bool:
    """
    Analyse whether a Python package follows a set of contracts, and report on the results.

    This function attempts to handle and report all exceptions, too.

    Args:
        config_filename:    the filename to use to parse user options.
        limit_to_contracts: if supplied, only lint the contracts with the supplied ids.
        cache_dir:          the directory to use for caching. Pass None to disable caching.
        is_debug_mode:      whether debugging should be turned on. In debug mode, exceptions are
                            not swallowed at the top level, so the stack trace can be seen.
        show_timings:       whether to show the times taken to build the graph and to check
                            each contract.
        verbose:            if True, noisily output progress as it goes along.

    Returns:
        True if the linting passed, False if it didn't.
    """
    rendering.print_title()

    output.verbose_print(verbose, "Verbose mode.")
    try:
        user_options = read_user_options(config_filename=config_filename)
        _register_contract_types(user_options)
        report = create_report(user_options, limit_to_contracts, cache_dir, show_timings, verbose)
    except Exception as e:
        if is_debug_mode:
            raise e
        render_exception(e)
        return FAILURE

    render_report(report)

    if report.contains_failures:
        return FAILURE
    else:
        return SUCCESS


def read_user_options(config_filename: str | None = None) -> UserOptions:
    """
    Return the UserOptions object from the supplied config file.

    If no filename is supplied, look in the default location
    (see importlinter.cli.lint_imports).

    Raises:
        FileNotFoundError if no configuration file could be found.
    """
    readers = settings.USER_OPTION_READERS.values()
    if config_filename:
        if config_filename.endswith(".toml"):
            readers = [settings.USER_OPTION_READERS["toml"]]
        else:
            readers = [settings.USER_OPTION_READERS["ini"]]

    for reader in readers:
        options = reader.read_options(config_filename=config_filename)
        if options:
            normalized_options = _normalize_user_options(options)
            return normalized_options
    raise FileNotFoundError("Could not read any configuration.")


def create_report(
    user_options: UserOptions,
    limit_to_contracts: tuple[str, ...] = tuple(),
    cache_dir: str | None | type[NotSupplied] = NotSupplied,
    show_timings: bool = False,
    verbose: bool = False,
) -> Report:
    """
    Analyse whether a Python package follows a set of contracts, returning a report on the results.

    Raises:
        InvalidUserOptions: if the report could not be run due to invalid user configuration,
                            such as a module that could not be imported.
    """
    include_external_packages = _get_include_external_packages(user_options)
    exclude_type_checking_imports = _get_exclude_type_checking_imports(user_options)

    with _get_spinner(":brick: Building graph...", verbose):
        with settings.TIMER as timer:
            graph = _build_graph(
                root_package_names=user_options.session_options["root_packages"],
                cache_dir=cache_dir,
                include_external_packages=include_external_packages,
                exclude_type_checking_imports=exclude_type_checking_imports,
                verbose=verbose,
            )
    graph_building_duration = timer.duration_in_ms

    output.verbose_print(verbose, f"Built graph in {format_duration(graph_building_duration)}.")

    return _build_report(
        graph=graph,
        graph_building_duration=graph_building_duration,
        user_options=user_options,
        limit_to_contracts=limit_to_contracts,
        show_timings=show_timings,
        verbose=verbose,
    )


# Private functions
# -----------------


def _normalize_user_options(user_options: UserOptions) -> UserOptions:
    normalized_options = copy(user_options)
    if "root_packages" not in normalized_options.session_options:
        normalized_options.session_options["root_packages"] = [
            normalized_options.session_options["root_package"]
        ]
    if "root_package" in normalized_options.session_options:
        del normalized_options.session_options["root_package"]
    return normalized_options


def _get_spinner(message: str, verbose: bool) -> contextlib.AbstractContextManager:
    if verbose:
        # The output in verbose mode interferes with the spinner so we disable it.
        return contextlib.nullcontext()
    else:
        return console.status(":brick: Building graph...", spinner="point")


def _build_graph(
    root_package_names: list[str],
    include_external_packages: bool | None,
    exclude_type_checking_imports: bool,
    verbose: bool,
    cache_dir: str | None | type[NotSupplied] = NotSupplied,
) -> ImportGraph:
    if cache_dir == NotSupplied:
        cache_dir = settings.DEFAULT_CACHE_DIR

    if cache_dir:
        output.verbose_print(verbose, f"Building import graph (cache directory is {cache_dir})...")
    else:
        output.verbose_print(verbose, "Building import graph (with caching disabled)...")

    return settings.GRAPH_BUILDER.build(
        root_package_names=root_package_names,
        include_external_packages=include_external_packages,
        exclude_type_checking_imports=exclude_type_checking_imports,
        cache_dir=cache_dir,
    )


def _build_report(
    graph: ImportGraph,
    graph_building_duration: int,
    user_options: UserOptions,
    limit_to_contracts: tuple[str, ...],
    show_timings: bool,
    verbose: bool,
) -> Report:
    report = Report(
        graph=graph,
        show_timings=show_timings,
        graph_building_duration=graph_building_duration,
    )
    contracts_options = _filter_contract_options(
        user_options.contracts_options, limit_to_contracts
    )

    contract_checking_progress = _get_contract_checking_progress(verbose)
    with contract_checking_progress:
        with Live(transient=True) as live:
            for contract_options in contract_checking_progress.track(contracts_options):
                contract_class = registry.get_contract_class(contract_options["type"])
                try:
                    contract = contract_class(
                        name=contract_options["name"],
                        session_options=user_options.session_options,
                        contract_options=contract_options,
                    )
                except InvalidContractOptions as e:
                    report.add_invalid_contract_options(contract_options["name"], e)
                    return report

                live.update(f"[dim]Checking {contract_options['name']}")
                output.verbose_print(verbose, f"Checking {contract.name}...")

                with settings.TIMER as timer:
                    # Make a copy so that contracts can mutate the graph without affecting
                    # other contract checks.
                    copy_of_graph = deepcopy(graph)
                    check = contract.check(copy_of_graph, verbose=verbose)
                duration_ms = timer.duration_in_ms
                report.add_contract_check(contract, check, duration=duration_ms)
                if verbose:
                    rendering.render_contract_result_line(contract, check, duration=duration_ms)

    output.verbose_print(verbose, newline=True)
    return report


def _get_contract_checking_progress(verbose: bool) -> Progress:
    return Progress(
        TextColumn(":face_with_monocle: Checking contracts"),
        BarColumn(),
        MofNCompleteColumn(),
        disable=verbose,
        transient=True,
    )


def _filter_contract_options(
    contracts_options: list[dict[str, Any]], limit_to_contracts: tuple[str, ...]
) -> list[dict[str, Any]]:
    if limit_to_contracts:
        # Validate the supplied contract ids.
        registered_contract_ids = {option["id"] for option in contracts_options}
        missing_contract_ids = set(limit_to_contracts) - registered_contract_ids
        if missing_contract_ids:
            if len(missing_contract_ids) == 1:
                raise ValueError(
                    f"Could not find contract '{missing_contract_ids.pop()}'.\n\n"
                    "You asked to limit the check to that contract, but nothing exists "
                    "with that id."
                )
            else:
                raise ValueError(
                    "Could not find the following contract ids: "
                    f"{', '.join(sorted(missing_contract_ids))}.\n\n"
                    "You asked to limit the check to those contracts, but there are no "
                    "contracts with those ids."
                )
        else:
            return [o for o in contracts_options if o["id"] in limit_to_contracts]
    else:
        return contracts_options


def _register_contract_types(user_options: UserOptions) -> None:
    contract_types = _get_built_in_contract_types() + _get_plugin_contract_types(user_options)
    for name, contract_class in contract_types:
        registry.register(contract_class, name)


def _get_built_in_contract_types() -> list[tuple[str, type[Contract]]]:
    return list(
        map(
            _parse_contract_type_string,
            [
                "forbidden: importlinter.contracts.forbidden.ForbiddenContract",
                "layers: importlinter.contracts.layers.LayersContract",
                "independence: importlinter.contracts.independence.IndependenceContract",
                "protected: importlinter.contracts.protected.ProtectedContract",
                "acyclic_siblings: importlinter.contracts.acyclic_siblings.AcyclicSiblingsContract",
            ],
        )
    )


def _get_plugin_contract_types(
    user_options: UserOptions,
) -> list[tuple[str, type[Contract]]]:
    contract_types = []
    if "contract_types" in user_options.session_options:
        for contract_type_string in user_options.session_options["contract_types"]:
            contract_types.append(_parse_contract_type_string(contract_type_string))
    return contract_types


def _parse_contract_type_string(string) -> tuple[str, type[Contract]]:
    components = string.split(": ")
    assert len(components) == 2
    name, contract_class_string = components
    contract_class = _string_to_class(contract_class_string)
    if not issubclass(contract_class, Contract):
        raise TypeError(f"{contract_class} is not a subclass of Contract.")
    return name, contract_class


def _string_to_class(string: str) -> type:
    """
    Parse a string into a Python class.

    Args:
        string: a fully qualified string of a class, e.g. 'mypackage.foo.MyClass'.

    Returns:
        The class.
    """
    components = string.split(".")
    class_name = components[-1]
    module_name = ".".join(components[:-1])
    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)
    assert isinstance(cls, type)
    return cls


def _get_include_external_packages(user_options: UserOptions) -> bool | None:
    """
    Get a boolean (or None) for the include_external_packages option in user_options.
    """
    try:
        include_external_packages_str = user_options.session_options["include_external_packages"]
    except KeyError:
        return None
    # Cast the string to a boolean.
    return include_external_packages_str in ("True", "true")


def _get_exclude_type_checking_imports(user_options: UserOptions) -> bool:
    """
    Get a boolean for the exclude_type_checking_imports option in user_options.
    """
    try:
        exclude_type_checking_imports_str = user_options.session_options[
            "exclude_type_checking_imports"
        ]
    except KeyError:
        return False
    # Cast the string to a boolean.
    return exclude_type_checking_imports_str in ("True", "true")


def _get_show_timings(user_options: UserOptions) -> bool:
    """
    Get a boolean (or None) for the show_timings option in user_options.
    """
    try:
        show_timings_str = user_options.session_options["show_timings"]
    except KeyError:
        return False
    # Cast the string to a boolean.
    return show_timings_str in ("True", "true")



from __future__ import annotations
from textwrap import dedent
import pytest
from grimp import ImportGraph
from importlinter.application.output import console
from importlinter.application.app_config import settings
from importlinter.contracts.independence import (
    IndependenceContract,
    _SubpackageChainData,
)
from importlinter.domain.contract import ContractCheck

from tests.adapters.timing import FakeTimer


@pytest.fixture(scope="module", autouse=True)
def configure():
    settings.configure(TIMER=FakeTimer())


class TestIndependenceContract:
    def _build_default_graph(self):
        graph = ImportGraph()
        for module in (
            "mypackage",
            "mypackage.blue",
            "mypackage.blue.alpha",
            "mypackage.blue.beta",
            "mypackage.blue.beta.foo",
            "mypackage.green",
            "mypackage.yellow",
            "mypackage.yellow.gamma",
            "mypackage.yellow.delta",
            "mypackage.other",
        ):
            graph.add_module(module)
        return graph

    def _check_default_contract(self, graph):
        contract = IndependenceContract(
            name="Independence contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "modules": ("mypackage.blue", "mypackage.green", "mypackage.yellow")
            },
        )
        return contract.check(graph=graph, verbose=False)

    def _check_wildcard_contract(self, graph):
        contract = IndependenceContract(
            name="Independence contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={"modules": ("mypackage.*",)},
        )
        return contract.check(graph=graph, verbose=False)

    def test_when_modules_are_independent(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue",
            imported="mypackage.other",
            line_number=10,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.green",
            imported="mypackage.other",
            line_number=11,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)
        assert contract_check.kept

    def test_when_wildcard_modules_are_independent(self):
        graph = self._build_default_graph()

        contract_check = self._check_wildcard_contract(graph)
        assert contract_check.kept

    def test_when_wildcard_modules_are_not_independent(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue",
            imported="mypackage.green",
            line_number=10,
            line_contents="-",
        )

        contract_check = self._check_wildcard_contract(graph)
        assert not contract_check.kept

    def test_when_root_imports_root_directly(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue",
            imported="mypackage.green",
            line_number=10,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.blue",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue",
                                    "imported": "mypackage.green",
                                    "line_numbers": (10,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                }
            ]
        }
        assert expected_metadata == contract_check.metadata

    def test_when_root_imports_root_indirectly(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue",
            imported="mypackage.other",
            line_number=10,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.other",
            imported="mypackage.green",
            line_number=11,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.blue",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue",
                                    "imported": "mypackage.other",
                                    "line_numbers": (10,),
                                },
                                {
                                    "importer": "mypackage.other",
                                    "imported": "mypackage.green",
                                    "line_numbers": (11,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                }
            ]
        }
        assert expected_metadata == contract_check.metadata

    def test_chains_via_other_independent_modules_are_not_included(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue",
            imported="mypackage.green",
            line_number=10,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.yellow",
            imported="mypackage.blue",
            line_number=11,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.blue",
                    "downstream_module": "mypackage.yellow",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.yellow",
                                    "imported": "mypackage.blue",
                                    "line_numbers": (11,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.blue",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue",
                                    "imported": "mypackage.green",
                                    "line_numbers": (10,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                },
            ]
        }
        assert expected_metadata == {
            "invalid_chains": _sort_invalid_chains(contract_check.metadata["invalid_chains"])
        }

    def test_when_child_imports_child(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue.alpha",
            imported="mypackage.yellow.gamma",
            line_number=5,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.yellow",
                    "downstream_module": "mypackage.blue",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue.alpha",
                                    "imported": "mypackage.yellow.gamma",
                                    "line_numbers": (5,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                }
            ]
        }
        assert expected_metadata == contract_check.metadata

    def test_when_grandchild_imports_root(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.blue.beta.foo",
            imported="mypackage.green",
            line_number=8,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.blue",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue.beta.foo",
                                    "imported": "mypackage.green",
                                    "line_numbers": (8,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                }
            ]
        }
        assert expected_metadata == contract_check.metadata

    def test_extra_firsts_and_lasts(self):
        graph = self._build_default_graph()
        graph.add_import(
            importer="mypackage.yellow.foo",
            imported="mypackage.green.bar",
            line_number=15,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.green.foo",
            imported="mypackage.orange.bar",
            line_number=15,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.orange.bar",
            imported="mypackage.purple",
            line_number=4,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.purple",
            imported="mypackage.blue.foobar",
            line_number=41,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.green.bar",
            imported="mypackage.orange.bar",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.green.bar",
            imported="mypackage.orange.bar",
            line_number=2,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.green.bar.beta",
            imported="mypackage.orange.bar",
            line_number=31,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.purple",
            imported="mypackage.blue",
            line_number=1,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.purple",
            imported="mypackage.blue.baz.alpha",
            line_number=3,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.purple",
            imported="mypackage.blue.baz.alpha",
            line_number=16,
            line_contents="-",
        )

        contract_check = self._check_default_contract(graph)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.blue",
                    "downstream_module": "mypackage.green",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.green.bar",
                                    "imported": "mypackage.orange.bar",
                                    "line_numbers": (1, 2),
                                },
                                {
                                    "importer": "mypackage.orange.bar",
                                    "imported": "mypackage.purple",
                                    "line_numbers": (4,),
                                },
                                {
                                    "importer": "mypackage.purple",
                                    "imported": "mypackage.blue",
                                    "line_numbers": (1,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.green.bar.beta",
                                    "imported": "mypackage.orange.bar",
                                    "line_numbers": (31,),
                                },
                                {
                                    "importer": "mypackage.green.foo",
                                    "imported": "mypackage.orange.bar",
                                    "line_numbers": (15,),
                                },
                            ],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.purple",
                                    "imported": "mypackage.blue.baz.alpha",
                                    "line_numbers": (3, 16),
                                },
                                {
                                    "importer": "mypackage.purple",
                                    "imported": "mypackage.blue.foobar",
                                    "line_numbers": (41,),
                                },
                            ],
                        }
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.yellow",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.yellow.foo",
                                    "imported": "mypackage.green.bar",
                                    "line_numbers": (15,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        }
                    ],
                },
            ]
        }
        assert expected_metadata == {
            "invalid_chains": _sort_invalid_chains(contract_check.metadata["invalid_chains"])
        }


@pytest.mark.parametrize(
    "ignore_imports, is_kept",
    (
        (["mypackage.a -> mypackage.irrelevant"], False),
        (["mypackage.a -> mypackage.indirect"], True),
        (["mypackage.indirect -> mypackage.b"], True),
        # Wildcards
        (["mypackage.a -> *.irrelevant"], False),
        (["mypackage.a -> **.irrelevant"], False),
        (["*.a -> *.indirect"], True),
        (["**.a -> **.indirect"], True),
        (["mypackage.* -> mypackage.b"], True),
        (["mypackage.** -> mypackage.b"], True),
    ),
)
def test_ignore_imports(ignore_imports, is_kept):
    graph = ImportGraph()
    graph.add_module("mypackage")
    graph.add_import(
        importer="mypackage.a",
        imported="mypackage.irrelevant",
        line_number=1,
        line_contents="-",
    )
    graph.add_import(
        importer="mypackage.a",
        imported="mypackage.indirect",
        line_number=1,
        line_contents="-",
    )
    graph.add_import(
        importer="mypackage.indirect",
        imported="mypackage.b",
        line_number=1,
        line_contents="-",
    )
    contract = IndependenceContract(
        name="Independence contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "modules": ("mypackage.a", "mypackage.b"),
            "ignore_imports": ignore_imports,
        },
    )

    contract_check = contract.check(graph=graph, verbose=False)

    assert is_kept == contract_check.kept


def test_ignore_imports_adds_warnings():
    graph = ImportGraph()
    graph.add_module("mypackage")
    graph.add_import(
        importer="mypackage.green",
        imported="mypackage.blue",
        line_number=1,
        line_contents="-",
    )
    contract = IndependenceContract(
        name="Independence contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "modules": ("mypackage.green", "mypackage.blue"),
            "ignore_imports": [
                "mypackage.green.* -> mypackage.blue",
                "mypackage.nonexistent -> mypackage.blue",
            ],
            "unmatched_ignore_imports_alerting": "warn",
        },
    )

    contract_check = contract.check(graph=graph, verbose=False)

    assert set(contract_check.warnings) == {
        "No matches for ignored import mypackage.green.* -> mypackage.blue.",
        "No matches for ignored import mypackage.nonexistent -> mypackage.blue.",
    }


def test_render_broken_contract():
    contract = IndependenceContract(
        name="Independence contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={"modules": ["mypackage.blue", "mypackage.green", "mypackage.yellow"]},
    )
    check = ContractCheck(
        kept=False,
        metadata={
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.yellow",
                    "downstream_module": "mypackage.blue",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue.foo",
                                    "imported": "mypackage.utils.red",
                                    "line_numbers": (16, 102),
                                },
                                {
                                    "importer": "mypackage.utils.red",
                                    "imported": "mypackage.utils.brown",
                                    "line_numbers": (1,),
                                },
                                {
                                    "importer": "mypackage.utils.brown",
                                    "imported": "mypackage.yellow.bar",
                                    "line_numbers": (3,),
                                },
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.blue.bar",
                                    "imported": "mypackage.yellow.baz",
                                    "line_numbers": (5,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        },
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.yellow",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.yellow.foo",
                                    "imported": "mypackage.green.bar",
                                    "line_numbers": (15,),
                                }
                            ],
                            "extra_firsts": [],
                            "extra_lasts": [],
                        }
                    ],
                },
                {
                    "upstream_module": "mypackage.orange",
                    "downstream_module": "mypackage.brown",
                    "chains": [
                        {
                            "chain": [
                                {
                                    "importer": "mypackage.brown.foo",
                                    "imported": "mypackage.green.bar",
                                    "line_numbers": (15,),
                                },
                                {
                                    "importer": "mypackage.green.bar",
                                    "imported": "mypackage.yellow.foo",
                                    "line_numbers": (4,),
                                },
                                {
                                    "importer": "mypackage.yellow.foo",
                                    "imported": "mypackage.orange.foobar",
                                    "line_numbers": (41,),
                                },
                            ],
                            "extra_firsts": [
                                {
                                    "importer": "mypackage.brown.bar.alpha",
                                    "imported": "mypackage.green.bar",
                                    "line_numbers": (1, 2),
                                },
                                {
                                    "importer": "mypackage.brown.bar.beta",
                                    "imported": "mypackage.green.bar",
                                    "line_numbers": (31,),
                                },
                            ],
                            "extra_lasts": [
                                {
                                    "importer": "mypackage.yellow.foo",
                                    "imported": "mypackage.orange.delta",
                                    "line_numbers": (1,),
                                },
                                {
                                    "importer": "mypackage.yellow.foo",
                                    "imported": "mypackage.orange.gamma",
                                    "line_numbers": (3, 16),
                                },
                            ],
                        }
                    ],
                },
            ]
        },
    )

    with console.capture() as capture:
        contract.render_broken_contract(check)

    assert capture.get() == dedent(
        """\
        mypackage.blue is not allowed to import mypackage.yellow:

        - mypackage.blue.foo -> mypackage.utils.red (l.16, l.102)
          mypackage.utils.red -> mypackage.utils.brown (l.1)
          mypackage.utils.brown -> mypackage.yellow.bar (l.3)

        - mypackage.blue.bar -> mypackage.yellow.baz (l.5)


        mypackage.yellow is not allowed to import mypackage.green:

        - mypackage.yellow.foo -> mypackage.green.bar (l.15)


        mypackage.brown is not allowed to import mypackage.orange:

        - mypackage.brown.foo (l.15)
          & mypackage.brown.bar.alpha (l.1, l.2)
          & mypackage.brown.bar.beta -> mypackage.green.bar (l.31)
          mypackage.green.bar -> mypackage.yellow.foo (l.4)
          mypackage.yellow.foo -> mypackage.orange.foobar (l.41)
                                  & mypackage.orange.delta (l.1)
                                  & mypackage.orange.gamma (l.3, l.16)


        """
    )


def test_missing_module():
    graph = ImportGraph()
    for module in ("mypackage", "mypackage.foo"):
        graph.add_module(module)

    contract = IndependenceContract(
        name="Independence contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={"modules": ["mypackage.foo", "mypackage.bar"]},
    )

    with pytest.raises(ValueError, match=("Module 'mypackage.bar' does not exist.")):
        contract.check(graph=graph, verbose=False)


def test_ignore_imports_tolerates_duplicates():
    graph = ImportGraph()
    graph.add_module("mypackage")
    graph.add_import(
        importer="mypackage.a", imported="mypackage.b", line_number=1, line_contents="-"
    )
    graph.add_import(
        importer="mypackage.a", imported="mypackage.c", line_number=2, line_contents="-"
    )
    contract = IndependenceContract(
        name="Independence contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "modules": ("mypackage.a", "mypackage.b"),
            "ignore_imports": [
                "mypackage.a -> mypackage.b",
                "mypackage.a -> mypackage.c",
                "mypackage.a -> mypackage.b",
            ],
        },
    )

    contract_check = contract.check(graph=graph, verbose=False)

    assert contract_check.kept


@pytest.mark.parametrize(
    "independent_modules, is_kept",
    (
        (("namespace.portionone.blue", "namespace.subnamespace.portiontwo.blue"), True),
        (
            ("namespace.portionone.blue", "namespace.subnamespace.portiontwo.green"),
            False,
        ),
        (
            (
                "namespace.subnamespace.portiontwo.blue",
                "namespace.subnamespace.portiontwo.green",
            ),
            False,
        ),
    ),
)
def test_namespace_packages(independent_modules, is_kept):
    graph = ImportGraph()
    for module in (
        "portionone",
        "portionone.blue",
        "subnamespace.portiontwo",
        "subnamespace.portiontwo.green",
        "subnamespace.portiontwo.blue",
    ):
        graph.add_module(f"namespace.{module}")
    # Add imports between portions to another.
    graph.add_import(
        importer="namespace.portionone.blue",
        imported="namespace.subnamespace.portiontwo.green",
        line_number=3,
        line_contents="-",
    )
    graph.add_import(
        importer="namespace.subnamespace.portiontwo.blue",
        imported="namespace.subnamespace.portiontwo.green",
        line_number=3,
        line_contents="-",
    )

    contract = IndependenceContract(
        name="Independence contract",
        session_options={
            "root_packages": [
                "namespace.portionone",
                "namespace.subnamespace.portiontwo",
            ]
        },
        contract_options={
            "modules": independent_modules,
        },
    )

    contract_check = contract.check(graph=graph, verbose=False)

    assert contract_check.kept == is_kept


def _sort_invalid_chains(
    invalid_chains: list[_SubpackageChainData],
) -> list[_SubpackageChainData]:
    return sorted(invalid_chains, key=lambda i: (i["upstream_module"], i["downstream_module"]))



from grimp import ImportGraph

from importlinter.application import output
from importlinter.domain import fields
from importlinter.domain.contract import Contract, ContractCheck


class AlwaysPassesContract(Contract):
    warnings = fields.ListField(subfield=fields.StringField(), required=False)

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        return ContractCheck(kept=True, warnings=self.warnings)  # type: ignore

    def render_broken_contract(self, check: "ContractCheck") -> None:
        # No need to implement, will never fail.
        raise NotImplementedError  # pragma: nocover


class AlwaysFailsContract(Contract):
    warnings = fields.ListField(subfield=fields.StringField(), required=False)

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        return ContractCheck(kept=False, warnings=self.warnings)  # type: ignore

    def render_broken_contract(self, check: "ContractCheck") -> None:
        output.print("This contract will always fail.")


class NoisyContract(Contract):
    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        output.verbose_print(verbose, "Hello from the noisy contract!")
        return ContractCheck(kept=True)

    def render_broken_contract(self, check: "ContractCheck") -> None:
        # No need to implement, will never fail.
        raise NotImplementedError  # pragma: nocover


class ForbiddenImportContract(Contract):
    """
    Contract that defines a single forbidden import between
    two modules.
    """

    importer = fields.ModuleField()
    imported = fields.ModuleField()

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        forbidden_import_details = graph.get_import_details(
            importer=self.importer.name,  # type: ignore[attr-defined]
            imported=self.imported.name,  # type: ignore[attr-defined]
        )
        import_exists = bool(forbidden_import_details)

        return ContractCheck(
            kept=not import_exists,
            metadata={"forbidden_import_details": forbidden_import_details},
        )

    def render_broken_contract(self, check: "ContractCheck") -> None:
        output.print(f"{self.importer} is not allowed to import {self.imported}:")
        output.print()
        for details in check.metadata["forbidden_import_details"]:
            line_number = details["line_number"]
            line_contents = details["line_contents"]
            output.indent_cursor()
            output.print(f"{self.importer}:{line_number}: {line_contents}")


class FieldsContract(Contract):
    single_field = fields.StringField()
    multiple_field = fields.ListField(subfield=fields.StringField())
    import_field = fields.ImportExpressionField()
    required_field = fields.StringField()  # Fields are required by default.

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        raise NotImplementedError

    def render_broken_contract(self, check: "ContractCheck") -> None:
        raise NotImplementedError


class MutationCheckContract(Contract):
    """
    Contract for checking that contracts can't mutate the graph for other contracts.

    It checks that there are a certain number of modules and imports in the graph, then adds
    an extra import containing two new modules. We can check two such contracts and the second one
    will fail, if the graph gets mutated by other contracts.
    """

    number_of_modules = fields.IntegerField()
    number_of_imports = fields.IntegerField()

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        number_of_modules = int(self.number_of_modules)  # type: ignore
        number_of_imports = int(self.number_of_imports)  # type: ignore
        if not all(
            [
                number_of_modules == len(graph.modules),
                number_of_imports == graph.count_imports(),
            ]
        ):
            raise RuntimeError("Contract was mutated.")

        # Mutate graph.
        graph.add_import(importer="added-by-contract-1", imported="added-by-contract-2")
        return ContractCheck(kept=True)

    def render_broken_contract(self, check: "ContractCheck") -> None:
        raise NotImplementedError



from __future__ import annotations

from types import TracebackType

from importlinter.application.ports.timing import Timer


class FakeTimer(Timer):
    ARBITRARY_SECONDS_SINCE_EPOCH = 1_000_000

    def __init__(self) -> None:
        super().__init__()
        self._current_time_ms = int(self.ARBITRARY_SECONDS_SINCE_EPOCH * 1000)
        self._tick_duration = 1
        self._increment = 0

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._tick()
        super().__exit__(exc_type, exc_val, exc_tb)

    def get_current_time_ms(self) -> int:
        return self._current_time_ms

    def setup(self, tick_duration: int, increment: int) -> None:
        self._tick_duration = tick_duration
        self._increment = increment

    def _tick(self) -> None:
        self._current_time_ms += self._tick_duration * 1000
        self._tick_duration += self._increment



from importlinter.application.user_options import UserOptions


def test_user_options_returns_false_on_unsupported_type_for_equality():
    options = UserOptions(session_options={}, contracts_options=[])
    result = options == 1
    assert result is False



import pytest

from importlinter.adapters.user_options import (
    IniFileUserOptionReader,
    TomlFileUserOptionReader,
)
from importlinter.application.app_config import settings
from importlinter.application.user_options import UserOptions
from tests.adapters.filesystem import FakeFileSystem


@pytest.mark.parametrize("filename", ("setup.cfg", ".importlinter"))
@pytest.mark.parametrize(
    "contents, expected_options",
    (
        (
            """
            [something]
            # This file has no import-linter section.
            foo = 1
            bar = hello
            """,
            None,
        ),
        (
            """
            [something]
            foo = 1
            bar = hello

            [importlinter]
            foo = hello
            bar = 999
            """,
            UserOptions(session_options={"foo": "hello", "bar": "999"}, contracts_options=[]),
        ),
        (
            """
            [importlinter]
            foo = hello

            [importlinter:contract:contract-one]
            name=Contract One
            key=value
            multiple_values=
                one
                two
                three
                foo.one -> foo.two

            [importlinter:contract:contract-two];
            name=Contract Two
            baz=3
            """,
            UserOptions(
                session_options={"foo": "hello"},
                contracts_options=[
                    {
                        "name": "Contract One",
                        "id": "contract-one",
                        "key": "value",
                        "multiple_values": [
                            "one",
                            "two",
                            "three",
                            "foo.one -> foo.two",
                        ],
                    },
                    {"name": "Contract Two", "id": "contract-two", "baz": "3"},
                ],
            ),
        ),
    ),
)
def test_ini_file_reader(filename, contents, expected_options):
    settings.configure(
        FILE_SYSTEM=FakeFileSystem(
            content_map={f"/path/to/folder/{filename}": contents},
            working_directory="/path/to/folder",
        )
    )

    options = IniFileUserOptionReader().read_options()

    assert expected_options == options


@pytest.mark.parametrize(
    "passed_filename, expected_foo_value",
    (
        (None, "green"),
        ("custom.ini", "blue"),
        ("deeper/custom.ini", "purple"),
        ("nonexistent.ini", FileNotFoundError()),
    ),
)
def test_respects_passed_filename(passed_filename, expected_foo_value):
    settings.configure(
        FILE_SYSTEM=FakeFileSystem(
            content_map={
                "/path/to/folder/.importlinter": """
                        [importlinter]
                        foo = green
                    """,
                "/path/to/folder/custom.ini": """
                        [importlinter]
                        foo = blue
                    """,
                "/path/to/folder/deeper/custom.ini": """
                        [importlinter]
                        foo = purple
                    """,
            },
            working_directory="/path/to/folder",
        )
    )
    expected_options = UserOptions(
        session_options={"foo": expected_foo_value}, contracts_options=[]
    )

    reader = IniFileUserOptionReader()

    if isinstance(expected_foo_value, Exception):
        with pytest.raises(
            expected_foo_value.__class__, match=f"Could not find {passed_filename}."
        ):
            reader.read_options(config_filename=passed_filename)
    else:
        options = reader.read_options(config_filename=passed_filename)
        assert expected_options == options


@pytest.mark.parametrize(
    "contents, expected_options",
    (
        (
            """
            [something]
            # This file has no import-linter section.
            foo = 1
            bar = "hello"
            """,
            None,
        ),
        (
            """
            [something]
            foo = 1
            bar = "hello"

            [tool.importlinter]
            foo = "hello"
            bar = 999
            """,
            UserOptions(session_options={"foo": "hello", "bar": 999}, contracts_options=[]),
        ),
        (
            """
            [tool.importlinter]
            foo = "hello"
            include_external_packages = true
            exclude_type_checking_imports = true

            [[tool.importlinter.contracts]]
            id = "contract-one"
            name = "Contract One"
            key = "value"
            multiple_values = [
                "one",
                "two",
                "three",
                "foo.one -> foo.two",
            ]

            [[tool.importlinter.contracts]]
            name = "Contract Two"
            baz = 3
            """,
            UserOptions(
                session_options={
                    "foo": "hello",
                    "include_external_packages": "True",
                    "exclude_type_checking_imports": "True",
                },
                contracts_options=[
                    {
                        "name": "Contract One",
                        "id": "contract-one",
                        "key": "value",
                        "multiple_values": [
                            "one",
                            "two",
                            "three",
                            "foo.one -> foo.two",
                        ],
                    },
                    {"name": "Contract Two", "baz": 3},
                ],
            ),
        ),
    ),
)
def test_toml_file_reader(contents, expected_options):
    settings.configure(
        FILE_SYSTEM=FakeFileSystem(
            content_map={"/path/to/folder/pyproject.toml": contents},
            working_directory="/path/to/folder",
        )
    )

    options = TomlFileUserOptionReader().read_options()
    assert expected_options == options



from __future__ import annotations
import grimp
from importlinter.application import rendering
import dataclasses
from importlinter.configuration import settings
from collections.abc import Set, Mapping
from typing import cast
from importlinter.domain.contract import Contract, ContractCheck
from importlinter.domain.helpers import module_expressions_to_modules
from importlinter.domain import fields
from importlinter.domain.imports import ModuleExpression
from importlinter.application import contract_utils, output
from importlinter.application.contract_utils import AlertLevel
from dataclasses import dataclass

DEFAULT_DEPTH = 10
MAX_DEPENDENCIES_TO_RENDER = 5

# (importer, imported)
Import = tuple[str, str]


@dataclass(frozen=True, order=True)
class PackageSummary:
    package: str
    dependencies: Set[Dependency]


@dataclass(frozen=True, order=True)
class Dependency:
    downstream: str
    upstream: str
    num_imports: int


class AcyclicSiblingsContract(Contract):
    type_name = "acyclic_siblings"

    ancestors = fields.SetField(subfield=fields.ModuleExpressionField())
    depth = fields.IntegerField(minimum=0, default=DEFAULT_DEPTH)
    skip_descendants = fields.SetField(subfield=fields.ModuleExpressionField(), required=False)
    ignore_imports = fields.SetField(subfield=fields.ImportExpressionField(), required=False)
    unmatched_ignore_imports_alerting = fields.EnumField(AlertLevel, default=AlertLevel.ERROR)

    def check(self, graph: grimp.ImportGraph, verbose: bool) -> ContractCheck:
        self._check_no_ancestors_and_descendant_expressions_are_the_same()
        ancestors = self._get_concrete_ancestors(graph)
        descendants_to_skip = self._get_concrete_skipped_descendants(graph)
        depth: int = self.depth  # type: ignore

        warnings = contract_utils.remove_ignored_imports(
            graph=graph,
            ignore_imports=self.ignore_imports,  # type: ignore
            unmatched_alerting=self.unmatched_ignore_imports_alerting,  # type: ignore
        )

        cycle_breakers_by_package: dict[str, set[Import]] = {}

        for ancestor in sorted(ancestors):
            cycle_breakers_by_package_for_ancestor = self._nominate_cycle_breakers_recursively(
                graph, ancestor, depth, descendants_to_skip, verbose
            )
            cycle_breakers_by_package.update(cycle_breakers_by_package_for_ancestor)

        return ContractCheck(
            kept=not bool(cycle_breakers_by_package),
            warnings=warnings,
            metadata={
                "cycle_breakers_by_package": cycle_breakers_by_package,
                "summaries": self._build_summaries(cycle_breakers_by_package),
            },
        )

    def _check_no_ancestors_and_descendant_expressions_are_the_same(self) -> None:
        if self.skip_descendants and self.ancestors:
            overlap = cast(set, self.ancestors) & cast(set, self.skip_descendants)
            if overlap:
                first_overlap = sorted(list(overlap))[0]
                assert isinstance(first_overlap, ModuleExpression)
                if first_overlap.has_wildcard_expression():
                    raise ValueError(
                        f"Cannot skip descendant '{first_overlap}' as the same expression is in ancestors."
                    )
                else:
                    raise ValueError(f"Cannot skip '{first_overlap}' as it is also an ancestor.")

    def _get_concrete_ancestors(self, graph: grimp.ImportGraph) -> set[str]:
        ancestor_modules = module_expressions_to_modules(
            graph,
            sorted(self.ancestors),  # type: ignore
            raise_if_unmatched=True,
        )
        ancestors = {m.name for m in ancestor_modules}

        return ancestors

    def _get_concrete_skipped_descendants(self, graph: grimp.ImportGraph) -> set[str]:
        if self.skip_descendants is None:
            return set()

        skipped = module_expressions_to_modules(
            graph,
            sorted(self.skip_descendants),  # type: ignore
            raise_if_unmatched=True,
        )
        return {m.name for m in skipped}

    def _nominate_cycle_breakers_recursively(
        self,
        graph: grimp.ImportGraph,
        ancestor: str,
        remaining_depth: int,
        descendants_to_skip: set[str],
        verbose: bool,
    ) -> dict[str, set[Import]]:
        cycle_breakers_by_package: dict[str, set[Import]] = {}
        children = graph.find_children(ancestor)
        if not children:
            # Stop drilling down.
            return cycle_breakers_by_package

        # Check for cycles, assuming there are at least two.
        # (You can't have a cycle with only one child.)
        if len(children) >= 2:
            output.verbose_print(
                verbose,
                f"Searching for cycles between children of {ancestor}...",
            )
            with settings.TIMER as timer:
                cycle_breakers = graph.nominate_cycle_breakers(ancestor)
            duration = rendering.format_duration(timer.duration_in_ms)
            if cycle_breakers:
                cycle_breakers_by_package[ancestor] = cycle_breakers
                num_cycle_breakers = len(cycle_breakers)
                pluralized_cycles = "cycle" if num_cycle_breakers == 1 else "cycles"
                output.verbose_print(
                    verbose, f"Found {num_cycle_breakers} {pluralized_cycles} in {duration}."
                )
            else:
                output.verbose_print(
                    verbose,
                    f"No cycles found ({duration}).",
                )

        if remaining_depth:
            for child in sorted(children - descendants_to_skip):
                if cycle_breakers_by_descendant := self._nominate_cycle_breakers_recursively(
                    graph, child, remaining_depth - 1, descendants_to_skip, verbose
                ):
                    cycle_breakers_by_package.update(cycle_breakers_by_descendant)

        return cycle_breakers_by_package

    def _build_summaries(
        self, cycle_breakers_by_package: Mapping[str, Set[Import]]
    ) -> set[PackageSummary]:
        summaries: set[PackageSummary] = set()
        for package, cycle_breakers in cycle_breakers_by_package.items():
            summaries.add(
                PackageSummary(package, self._build_dependencies(package, cycle_breakers))
            )
        return summaries

    def _build_dependencies(
        self, package: str, cycle_breakers: Set[Import]
    ) -> frozenset[Dependency]:
        dependencies_by_child_pairs: dict[Import, Dependency] = {}
        package_components = package.split(".")
        num_dependency_components = len(package_components) + 1
        for importer, imported in cycle_breakers:
            downstream_child = ".".join(importer.split(".")[:num_dependency_components])
            upstream_child = ".".join(imported.split(".")[:num_dependency_components])
            child_pair = (downstream_child, upstream_child)
            try:
                existing_dependency = dependencies_by_child_pairs[child_pair]
                dependency = dataclasses.replace(
                    existing_dependency, num_imports=existing_dependency.num_imports + 1
                )
            except KeyError:
                dependency = Dependency(
                    downstream=downstream_child,
                    upstream=upstream_child,
                    num_imports=1,
                )
            dependencies_by_child_pairs[child_pair] = dependency

        return frozenset(dependencies_by_child_pairs.values())

    def render_broken_contract(self, check: ContractCheck) -> None:
        for summary in sorted(check.metadata["summaries"]):
            output.print_error(f"No cycles are allowed in {summary.package}.")
            num_dependencies = len(summary.dependencies)
            pluralized_dependencies = "dependencies" if num_dependencies != 1 else "dependency"
            output.print_error(
                f"It could be made acyclic by removing {num_dependencies} {pluralized_dependencies}:"
            )
            output.new_line()

            for dependency in sorted(summary.dependencies)[:MAX_DEPENDENCIES_TO_RENDER]:
                relative_downstream = dependency.downstream.split(".")[-1]
                relative_upstream = dependency.upstream.split(".")[-1]
                pluralized_import = "imports" if dependency.num_imports != 1 else "import"
                output.print_error(
                    f"- .{relative_downstream} -> .{relative_upstream} ({dependency.num_imports} {pluralized_import})"
                )

            num_extra = max(len(summary.dependencies) - MAX_DEPENDENCIES_TO_RENDER, 0)
            if num_extra:
                output.print_error(f"(and {num_extra} more).")
            output.new_line()



import pytest
from grimp import ImportGraph

from importlinter.application.contract_utils import AlertLevel, remove_ignored_imports
from importlinter.domain.helpers import MissingImport
from importlinter.domain.imports import (
    DirectImport,
    ImportExpression,
    Module,
    ModuleExpression,
)


class TestRemoveIgnoredImports:
    DIRECT_IMPORTS = [
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.yellow"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.purple"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(
            importer=Module("mypackage.green"),
            imported=Module("mypackage.blue"),
            line_number=1,
            line_contents="-",
        ),
        DirectImport(  # Direct Imports can appear twice, for different line numbers.
            importer=Module("mypackage.green"),
            imported=Module("mypackage.blue"),
            line_number=2,
            line_contents="-",
        ),
    ]

    @pytest.mark.parametrize("alert_level", [AlertLevel.NONE, AlertLevel.WARN, AlertLevel.ERROR])
    def test_no_unresolved_import_expressions(self, alert_level):
        graph = self._build_graph(self.DIRECT_IMPORTS)

        warnings = remove_ignored_imports(
            graph=graph,
            ignore_imports=[
                ImportExpression(
                    importer=ModuleExpression("mypackage.green"),
                    imported=ModuleExpression("mypackage.blue"),
                ),
                ImportExpression(
                    importer=ModuleExpression("mypackage.green"),
                    imported=ModuleExpression("mypackage.purple"),
                ),
            ],
            unmatched_alerting=alert_level,
        )

        assert graph.count_imports() == 1  # The three matching imports have been removed.
        assert warnings == []

    @pytest.mark.parametrize(
        "alert_level, expected_result",
        [
            (AlertLevel.NONE, []),
            (
                AlertLevel.WARN,
                [
                    "No matches for ignored import mypackage.* -> mypackage.nonexistent.",
                    "No matches for ignored import mypackage.nonexistent -> mypackage.blue.",
                ],
            ),
            (
                AlertLevel.ERROR,
                MissingImport(
                    "No matches for ignored import mypackage.* -> mypackage.nonexistent."
                ),
            ),
        ],
    )
    def test_unresolved_import_expressions(self, alert_level, expected_result):
        graph = self._build_graph(self.DIRECT_IMPORTS)
        ignore_imports = [
            ImportExpression(
                importer=ModuleExpression("mypackage.green"),
                imported=ModuleExpression("mypackage.blue"),
            ),
            ImportExpression(
                importer=ModuleExpression("mypackage.*"),
                imported=ModuleExpression("mypackage.nonexistent"),
            ),
            ImportExpression(
                importer=ModuleExpression("mypackage.green"),
                imported=ModuleExpression("mypackage.purple"),
            ),
            ImportExpression(
                importer=ModuleExpression("mypackage.nonexistent"),
                imported=ModuleExpression("mypackage.blue"),
            ),
        ]

        if isinstance(expected_result, Exception):
            with pytest.raises(type(expected_result), match=str(expected_result)):
                remove_ignored_imports(
                    graph=graph,
                    ignore_imports=ignore_imports,
                    unmatched_alerting=alert_level,
                )
        else:
            warnings = remove_ignored_imports(
                graph=graph,
                ignore_imports=ignore_imports,
                unmatched_alerting=alert_level,
            )
            assert graph.count_imports() == 1  # The three matching imports have been removed.
            assert set(warnings) == set(expected_result)

    def _build_graph(self, direct_imports):
        graph = ImportGraph()
        for direct_import in direct_imports:
            graph.add_import(
                importer=direct_import.importer.name,
                imported=direct_import.imported.name,
                line_number=direct_import.line_number,
                line_contents=direct_import.line_contents,
            )
        return graph



import pytest

from importlinter.application import file_finding
from importlinter.application.app_config import settings
from tests.adapters.filesystem import FakeFileSystem


@pytest.mark.parametrize(
    "filenames, expected_result",
    (
        (["foo.txt"], ["/path/to/folder/foo.txt"]),
        (
            ["foo.txt", ".another"],
            ["/path/to/folder/foo.txt", "/path/to/folder/.another"],
        ),
        (["bar.txt"], []),
    ),
)
def test_finds_file_in_current_directory(filenames, expected_result):
    settings.configure(
        FILE_SYSTEM=FakeFileSystem(
            """
                /path/to/folder/
                    foo.txt
                    .another
                    another/
                        foo.txt
                        bar.txt
            """,
            working_directory="/path/to/folder",
        )
    )

    result = file_finding.find_any(*filenames)

    assert expected_result == result



import pytest
from grimp import ImportGraph

from textwrap import dedent
from importlinter.application.output import console
from importlinter.configuration import settings
from importlinter.contracts.forbidden import ForbiddenContract
from importlinter.domain.contract import ContractCheck
from tests.adapters.timing import FakeTimer


@pytest.fixture(scope="module", autouse=True)
def configure():
    settings.configure(TIMER=FakeTimer())


class TestForbiddenContract:
    @pytest.mark.parametrize("as_packages", ("False", "True"))
    def test_is_kept_when_no_forbidden_modules_imported(
        self,
        as_packages: bool,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.blue", "mypackage.yellow"),
            as_packages=as_packages,
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept

    @pytest.mark.parametrize(
        "as_packages, forbidden_modules, expected_invalid_chains",
        [
            (
                True,
                (
                    "mypackage.blue",
                    "mypackage.green",
                    "mypackage.yellow",
                    "mypackage.purple",
                ),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.one",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.one.alpha",
                                    "imported": "mypackage.green.beta",
                                    "line_numbers": (3,),
                                }
                            ],
                            [
                                {
                                    "importer": "mypackage.one.alpha.circle",
                                    "imported": "mypackage.green.beta.sphere",
                                    "line_numbers": (8,),
                                },
                            ],
                        ],
                    },
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                }
                            ]
                        ],
                    },
                    {
                        "upstream_module": "mypackage.purple",
                        "downstream_module": "mypackage.two",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.two",
                                    "imported": "mypackage.utils",
                                    "line_numbers": (9,),
                                },
                                {
                                    "importer": "mypackage.utils",
                                    "imported": "mypackage.purple",
                                    "line_numbers": (1,),
                                },
                            ]
                        ],
                    },
                ],
            ),
            (
                False,
                (
                    "mypackage.blue",
                    "mypackage.green",
                    "mypackage.yellow",
                    "mypackage.purple",
                ),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                }
                            ]
                        ],
                    },
                    {
                        "upstream_module": "mypackage.purple",
                        "downstream_module": "mypackage.two",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.two",
                                    "imported": "mypackage.utils",
                                    "line_numbers": (9,),
                                },
                                {
                                    "importer": "mypackage.utils",
                                    "imported": "mypackage.purple",
                                    "line_numbers": (1,),
                                },
                            ]
                        ],
                    },
                ],
            ),
        ],
    )
    def test_is_broken_when_forbidden_modules_imported(
        self,
        as_packages: str,
        forbidden_modules: tuple[str],
        expected_invalid_chains: list[dict],
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=forbidden_modules,
            as_packages=as_packages,
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert not contract_check.kept

        expected_metadata = {"invalid_chains": expected_invalid_chains}

        assert expected_metadata == contract_check.metadata

    @pytest.mark.parametrize(
        "as_packages",
        (False, True),
    )
    def test_is_broken_when_forbidden_external_modules_imported(
        self,
        as_packages: str,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("sqlalchemy", "requests"),
            include_external_packages=True,
            as_packages=as_packages,
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert not contract_check.kept

        expected_metadata = {
            "invalid_chains": [
                {
                    "upstream_module": "sqlalchemy",
                    "downstream_module": "mypackage.three",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.three",
                                "imported": "sqlalchemy",
                                "line_numbers": (1,),
                            }
                        ]
                    ],
                }
            ]
        }

        assert expected_metadata == contract_check.metadata

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    def test_is_invalid_when_forbidden_externals_but_graph_does_not_include_externals(
        self,
        as_packages: bool,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("sqlalchemy", "requests"),
            as_packages=as_packages,
        )

        with pytest.raises(
            ValueError,
            match=(
                "The top level configuration must have include_external_packages=True "
                "when there are external forbidden modules."
            ),
        ):
            contract.check(graph=graph, verbose=False)

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    def test_ignore_imports_tolerates_duplicates(self, as_packages: bool):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.blue", "mypackage.yellow"),
            ignore_imports=(
                "mypackage.three -> mypackage.green",
                "mypackage.utils -> mypackage.purple",
                "mypackage.three -> mypackage.green",
            ),
            include_external_packages=False,
            as_packages=as_packages,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.kept

    def test_wildcards_in_source_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green"),
            source_modules=("mypackage.one.*",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.one.alpha",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    def test_recursive_wildcards_in_source_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green"),
            source_modules=("mypackage.one.**",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.one.alpha",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.one.alpha.circle",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    def test_wildcards_in_forbidden_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green.*"),
            source_modules=("mypackage.one",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green.beta",
                    "downstream_module": "mypackage.one",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    def test_recursive_wildcards_in_forbidden_modules_are_resolved(self):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green.**"),
            source_modules=("mypackage.one",),
            include_external_packages=False,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green.beta",
                    "downstream_module": "mypackage.one",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha",
                                "imported": "mypackage.green.beta",
                                "line_numbers": (3,),
                            },
                        ],
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
                {
                    "upstream_module": "mypackage.green.beta.sphere",
                    "downstream_module": "mypackage.one",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.one.alpha.circle",
                                "imported": "mypackage.green.beta.sphere",
                                "line_numbers": (8,),
                            },
                        ],
                    ],
                },
            ],
        }

    @pytest.mark.parametrize(
        "as_packages, forbidden_modules, expected_invalid_chains",
        [
            (
                True,
                ("mypackage.green",),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.one",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.one.alpha.circle",
                                    "imported": "mypackage.green.beta.sphere",
                                    "line_numbers": (8,),
                                },
                            ]
                        ],
                    },
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                },
                            ]
                        ],
                    },
                ],
            ),
            (
                False,
                ("mypackage.green",),
                [
                    {
                        "upstream_module": "mypackage.green",
                        "downstream_module": "mypackage.three",
                        "chains": [
                            [
                                {
                                    "importer": "mypackage.three",
                                    "imported": "mypackage.green",
                                    "line_numbers": (4,),
                                },
                            ]
                        ],
                    },
                ],
            ),
        ],
    )
    def test_ignore_imports_with_wildcards(
        self,
        as_packages: str,
        forbidden_modules: tuple[str],
        expected_invalid_chains: list[dict],
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=forbidden_modules,
            ignore_imports=("mypackage.*.alpha -> mypackage.*.beta",),
            as_packages=as_packages,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": expected_invalid_chains,
        }

    @pytest.mark.parametrize("as_packages", (False, True))
    def test_ignore_imports_with_recursive_wildcards(
        self,
        as_packages: str,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.green",),
            ignore_imports=(
                "mypackage.**.alpha -> mypackage.**.beta",
                "mypackage.**.circle -> mypackage.**.sphere",
            ),
            as_packages=as_packages,
        )

        check = contract.check(graph=graph, verbose=False)
        assert check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.three",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.three",
                                "imported": "mypackage.green",
                                "line_numbers": (4,),
                            }
                        ]
                    ],
                },
            ]
        }

    @pytest.mark.parametrize(
        "importer",
        ("mypackage.one", "mypackage.one.alpha"),
    )
    @pytest.mark.parametrize(
        "imported",
        ("mypackage.mauve", "mypackage.mauve.beta"),
    )
    @pytest.mark.parametrize(
        "allow_indirect_imports",
        (False, True),
    )
    def test_allow_indirect_imports(self, importer, imported, allow_indirect_imports):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.mauve",),
            allow_indirect_imports=allow_indirect_imports,
        )
        graph.add_module("mypackage.mauve")
        # Add a direct import.
        graph.add_import(
            importer=importer,
            imported=imported,
            line_number=10,
            line_contents="-",
        )
        # Add an indirect import.
        graph.add_import(
            importer="mypackage.one.delta",
            imported="mypackage.something",
            line_number=20,
            line_contents="-",
        )
        graph.add_import(
            importer="mypackage.something",
            imported="mypackage.mauve.gamma",
            line_number=30,
            line_contents="-",
        )

        contract_check = contract.check(graph=graph, verbose=False)

        direct_chain = [
            {
                "importer": importer,
                "imported": imported,
                "line_numbers": (10,),
            },
        ]
        indirect_chain = [
            {
                "importer": "mypackage.one.delta",
                "imported": "mypackage.something",
                "line_numbers": (20,),
            },
            {
                "importer": "mypackage.something",
                "imported": "mypackage.mauve.gamma",
                "line_numbers": (30,),
            },
        ]
        if allow_indirect_imports:
            expected_chains = [direct_chain]
        else:
            expected_chains = [direct_chain, indirect_chain]
        assert contract_check.metadata == {
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.mauve",
                    "downstream_module": "mypackage.one",
                    "chains": expected_chains,
                },
            ],
        }

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    def test_ignore_imports_adds_warnings(self, as_packages: bool):
        graph = self._build_graph()
        contract = ForbiddenContract(
            name="Forbid contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "source_modules": ("mypackage.one", "mypackage.two", "mypackage.three"),
                "forbidden_modules": "mypackage.purple",
                "ignore_imports": [
                    "mypackage.one -> mypackage.two.nonexistent",
                    "mypackage.*.nonexistent -> mypackage.three",
                ],
                "unmatched_ignore_imports_alerting": "warn",
                "as_packages": as_packages,
            },
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert set(contract_check.warnings) == {
            "No matches for ignored import mypackage.one -> mypackage.two.nonexistent.",
            "No matches for ignored import mypackage.*.nonexistent -> mypackage.three.",
        }

    @pytest.mark.parametrize(
        "as_packages",
        (
            ("False"),
            ("True"),
        ),
    )
    @pytest.mark.parametrize(
        "module, expected_error",
        (
            (
                "requests.something",
                (
                    "Invalid forbidden module requests.something: "
                    "subpackages of external packages are not valid."
                ),
            ),
            # N.B. google.protobuf is a namespace package, but unless it is specified
            # as a root package, it will just appear in the graph as an external package
            # named 'google', so won't be treated any differently to other packages.
            (
                "google.protobuf",
                (
                    "Invalid forbidden module google.protobuf: "
                    "subpackages of external packages are not valid."
                ),
            ),
        ),
    )
    def test_is_invalid_when_subpackages_of_external_packages_are_provided(
        self,
        module: str,
        expected_error: str,
        as_packages: bool,
    ):
        graph = self._build_graph()
        contract = self._build_contract(
            forbidden_modules=("mypackage.blue", module),
            include_external_packages=True,
            as_packages=as_packages,
        )

        with pytest.raises(
            ValueError,
            match=expected_error,
        ):
            contract.check(graph=graph, verbose=False)

    def _build_graph(self):
        graph = ImportGraph()
        for module in (
            "one",
            "one.alpha",
            "one.alpha.circle",
            "two",
            "three",
            "blue",
            "green",
            "green.beta",
            "green.beta.sphere",
            "yellow",
            "purple",
            "utils",
        ):
            graph.add_module(f"mypackage.{module}")
        for external_module in ("sqlalchemy", "requests", "google"):
            graph.add_module(external_module, is_squashed=True)
        graph.add_import(
            importer="mypackage.one.alpha",
            imported="mypackage.green.beta",
            line_number=3,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.one.alpha.circle",
            imported="mypackage.green.beta.sphere",
            line_number=8,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="mypackage.green",
            line_number=4,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.two",
            imported="mypackage.utils",
            line_number=9,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.utils",
            imported="mypackage.purple",
            line_number=1,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="sqlalchemy",
            line_number=1,
            line_contents="foo",
        )
        return graph

    def _build_contract(
        self,
        forbidden_modules,
        ignore_imports=None,
        include_external_packages=False,
        allow_indirect_imports=None,
        source_modules=None,
        as_packages=True,
    ):
        session_options = {"root_packages": ["mypackage"]}
        if include_external_packages:
            session_options["include_external_packages"] = "True"

        contract_options = {
            "source_modules": source_modules
            or ("mypackage.one", "mypackage.two", "mypackage.three"),
            "forbidden_modules": forbidden_modules,
            "ignore_imports": ignore_imports or [],
        }

        if allow_indirect_imports is not None:
            contract_options["allow_indirect_imports"] = (
                "true" if allow_indirect_imports else "false"
            )
        contract_options["as_packages"] = "true" if as_packages else "false"

        return ForbiddenContract(
            name="Forbid contract",
            session_options=session_options,
            contract_options=contract_options,
        )


class TestForbiddenContractForNamespacePackages:
    @pytest.mark.parametrize(
        "forbidden, is_kept",
        [
            ("namespace.subnamespace.portiontwo.green", False),
            ("namespace.subnamespace.portiontwo.blue", True),
        ],
    )
    def test_allows_forbidding_of_inter_portion_imports(self, forbidden, is_kept):
        graph = ImportGraph()
        for module in (
            "portionone",
            "portionone.blue",
            "subnamespace.portiontwo",
            "subnamespace.portiontwo.green",
            "subnamespace.portiontwo.blue",
        ):
            graph.add_module(f"namespace.{module}")
        for external_module in ("sqlalchemy", "requests"):
            graph.add_module(external_module, is_squashed=True)
        # Add import from one portion to another.
        graph.add_import(
            importer="namespace.portionone.blue",
            imported="namespace.subnamespace.portiontwo.green",
            line_number=3,
            line_contents="-",
        )
        contract = self._build_contract(
            root_packages=["namespace.portionone", "namespace.subnamespace.portiontwo"],
            source_modules=("namespace.portionone.blue",),
            forbidden_modules=(forbidden,),
        )

        contract_check = contract.check(graph=graph, verbose=False)

        assert contract_check.kept == is_kept

    def _build_contract(
        self,
        root_packages,
        source_modules,
        forbidden_modules,
        include_external_packages=False,
    ):
        session_options = {"root_packages": root_packages}
        if include_external_packages:
            session_options["include_external_packages"] = "True"

        return ForbiddenContract(
            name="Forbid contract",
            session_options=session_options,
            contract_options={
                "source_modules": source_modules,
                "forbidden_modules": forbidden_modules,
            },
        )


def test_render_broken_contract():
    contract = ForbiddenContract(
        name="Forbid contract",
        session_options={"root_packages": ["mypackage"]},
        contract_options={
            "source_modules": ("mypackage.one", "mypackage.two", "mypackage.three"),
            "forbidden_modules": (
                "mypackage.blue",
                "mypackage.green",
                "mypackage.yellow",
                "mypackage.purple",
            ),
        },
    )
    check = ContractCheck(
        kept=False,
        metadata={
            "invalid_chains": [
                {
                    "upstream_module": "mypackage.purple",
                    "downstream_module": "mypackage.two",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.two",
                                "imported": "mypackage.utils",
                                "line_numbers": (9,),
                            },
                            {
                                "importer": "mypackage.utils",
                                "imported": "mypackage.purple",
                                "line_numbers": (1,),
                            },
                        ]
                    ],
                },
                {
                    "upstream_module": "mypackage.green",
                    "downstream_module": "mypackage.three",
                    "chains": [
                        [
                            {
                                "importer": "mypackage.three",
                                "imported": "mypackage.green",
                                "line_numbers": (4,),
                            }
                        ]
                    ],
                },
            ]
        },
    )

    with console.capture() as capture:
        contract.render_broken_contract(check)

    assert capture.get() == dedent(
        """\
        mypackage.two is not allowed to import mypackage.purple:

        -   mypackage.two -> mypackage.utils (l.9)
            mypackage.utils -> mypackage.purple (l.1)


        mypackage.three is not allowed to import mypackage.green:

        -   mypackage.three -> mypackage.green (l.4)


        """
    )


class TestVerbosePrint:
    def test_verbose(self):
        timer = FakeTimer()
        timer.setup(tick_duration=10, increment=0)
        settings.configure(TIMER=timer)

        graph = ImportGraph()
        for module in (
            "one",
            "one.alpha",
            "two",
            "three",
            "blue",
            "green",
            "green.beta",
            "yellow",
            "purple",
            "utils",
        ):
            graph.add_module(f"mypackage.{module}")
        graph.add_import(
            importer="mypackage.one.alpha",
            imported="mypackage.green.beta",
            line_number=3,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="mypackage.green",
            line_number=4,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.two",
            imported="mypackage.utils",
            line_number=9,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.utils",
            imported="mypackage.purple",
            line_number=1,
            line_contents="foo",
        )
        graph.add_import(
            importer="mypackage.three",
            imported="sqlalchemy",
            line_number=1,
            line_contents="foo",
        )
        contract = ForbiddenContract(
            name="Forbid contract",
            session_options={"root_packages": ["mypackage"]},
            contract_options={
                "source_modules": ("mypackage.one", "mypackage.two", "mypackage.three"),
                "forbidden_modules": (
                    "mypackage.blue",
                    "mypackage.green",
                    "mypackage.yellow",
                    "mypackage.purple",
                ),
                "ignore_imports": [],
                "allow_indirect_imports": "false",
            },
        )

        with console.capture() as capture:
            contract.check(graph=graph, verbose=True)

        assert capture.get() == dedent(
            """\
            Searching for import chains from mypackage.one to mypackage.blue...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.one to mypackage.green...
            Found 1 illegal chain in 10s.
            Searching for import chains from mypackage.one to mypackage.purple...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.one to mypackage.yellow...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.three to mypackage.blue...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.three to mypackage.green...
            Found 1 illegal chain in 10s.
            Searching for import chains from mypackage.three to mypackage.purple...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.three to mypackage.yellow...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.two to mypackage.blue...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.two to mypackage.green...
            Found 0 illegal chains in 10s.
            Searching for import chains from mypackage.two to mypackage.purple...
            Found 1 illegal chain in 10s.
            Searching for import chains from mypackage.two to mypackage.yellow...
            Found 0 illegal chains in 10s.
            """
        )



import os
import tempfile
from pathlib import Path

from importlinter import cli

this_directory = Path(__file__).parent
assets_directory = this_directory / ".." / "assets"

testpackage_directory = assets_directory / "testpackage"


class TestCaching:
    def test_no_cache(self):
        os.chdir(testpackage_directory)

        with tempfile.TemporaryDirectory() as cache_dir:
            cli.lint_imports(cache_dir=cache_dir, no_cache=True)

            meta_file = Path(cache_dir) / "testpackage.meta.json"
            # Blake2B 20-character hash of "testpackage".
            data_file = Path(cache_dir) / "216a42e0de035f511ad815351348f07aa91057c2.data.json"

            assert not meta_file.exists()
            assert not data_file.exists()

    def test_supplied_cache_dir(self):
        os.chdir(testpackage_directory)

        with tempfile.TemporaryDirectory() as cache_dir:
            cli.lint_imports(cache_dir=cache_dir, is_debug_mode=True)

            meta_file = Path(cache_dir) / "testpackage.meta.json"
            # Blake2B 20-character hash of "testpackage".
            data_file = Path(cache_dir) / "216a42e0de035f511ad815351348f07aa91057c2.data.json"

            assert meta_file.exists()
            assert data_file.exists()



from rich.console import Console


ERROR = "error"
SUCCESS = "success"
WARNING = "warning"
COLORS = {ERROR: "red", SUCCESS: "green", WARNING: "yellow"}

HEADING_LEVEL_ONE = 1
HEADING_LEVEL_TWO = 2
HEADING_LEVEL_THREE = 3

HEADING_MAP = {
    HEADING_LEVEL_ONE: ("=", True),
    HEADING_LEVEL_TWO: ("-", True),
    HEADING_LEVEL_THREE: ("-", False),
}

INDENT_SIZE = 4


console = Console(highlight=False)


class Output:
    """
    A class for writing output to the console.

    Output provides a few convenience methods for printing.

    For finer-grained control or more advanced user interface features,
    use the Rich Console object instantiated in this module (importlinter.application.output.console).
    """

    def print(
        self,
        text: str = "",
        bold: bool = False,
        color: str | None = None,
        newline: bool = True,
    ) -> None:
        """
        Print a line.

        Args:
            text (str):               The text to print.
            bold (bool, optional):    Whether to style the text in bold. (Default False.)
            color (str, optional):    The color of text to use. One of the values of the
                                      COLORS dictionary.
            newline (bool, optional): Whether to include a new line after the text.
                                      (Default True.)
        """
        adjusted_text = text
        if color:
            adjusted_text = f"[{color}]{text}[/{color}]"
        if bold:
            adjusted_text = f"[bold]{text}[/bold]"
        end = "\n" if newline else ""
        console.print(adjusted_text, end=end)

    def indent_cursor(self) -> None:
        """
        Indents the cursor ready to print a line.
        """
        console.print(" " * INDENT_SIZE, end="")

    def new_line(self) -> None:
        """
        Print a blank line.
        """
        console.print()

    def print_heading(self, text: str, level: int, style: str | None = None) -> None:
        """
        Prints the supplied text to the console, formatted as a heading.

        Args:
            text (str):            The text to format as a heading.
            level (int):           The level of heading to display (one of the keys
                                   of HEADING_MAP).
            style (str, optional): ERROR or SUCCESS style to apply (default None).
        Usage:

            output.print_heading('Foo', output.HEADING_LEVEL_ONE)
        """
        # Setup styling variables.
        is_bold = True
        color = COLORS[style] if style else None
        line_char, show_line_above = HEADING_MAP[level]
        heading_line = line_char * len(text)

        # Print lines.
        if show_line_above:
            self.print(heading_line, bold=is_bold, color=color)
        self.print(text, bold=is_bold, color=color)
        self.print(heading_line, bold=is_bold, color=color)
        self.print()

    def print_success(self, text: str, bold: bool = True) -> None:
        """
        Prints a line to the console, formatted as a success.
        """
        self.print(text, color=COLORS[SUCCESS], bold=bold)

    def print_error(self, text: str, bold: bool = True) -> None:
        """
        Prints a line to the console, formatted as an error.
        """
        self.print(text, color=COLORS[ERROR], bold=bold)

    def print_warning(self, text: str) -> None:
        """
        Prints a line to the console, formatted as a warning.
        """
        self.print(text, color=COLORS[WARNING])


# Use prebound method pattern to provide a simple API.
# https://python-patterns.guide/python/prebound-methods/
_instance = Output()
print = _instance.print
indent_cursor = _instance.indent_cursor
new_line = _instance.new_line
print_success = _instance.print_success
print_heading = _instance.print_heading
print_error = _instance.print_error
print_warning = _instance.print_warning


def verbose_print(
    verbose: bool,
    text: str = "",
    bold: bool = False,
    color: str | None = None,
    newline: bool = True,
) -> None:
    """
    Print a message, but only if we're in verbose mode.
    """
    if verbose:
        print(text, bold, color, newline)
